# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.watchers import bgp_watcher
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests import utils

CONF = cfg.CONF


class TestPortBindingChassisCreatedEvent(test_base.TestCase):

    def setUp(self):
        super(TestPortBindingChassisCreatedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = bgp_watcher.PortBindingChassisCreatedEvent(self.agent)

    def test_match_fn(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               up=[True])
        old = utils.create_row(chassis=[])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_not_single_or_dual_stack(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_old_chassis_set(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               up=[True])
        old = utils.create_row(chassis=[ch])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_no_old_chassis(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               up=[True])
        old = utils.create_row(chassis=[])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_different_old_chassis(self):
        ch = utils.create_row(name=self.chassis)
        ch_old = utils.create_row(name='old-chassis')
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               up=[True])
        old = utils.create_row(chassis=[ch_old])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_no_up(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock))

    def test_match_fn_no_old_up(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               up=[True])
        old = utils.create_row(chassis=[ch],
                               up=[False])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_index_error(self):
        row = utils.create_row(chassis=[],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_run(self):
        row = utils.create_row(type=constants.OVN_VIF_PORT_TYPES[0],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ip.assert_called_once_with(['10.10.1.16'], row)

    def test_run_dual_stack(self):
        row = utils.create_row(
            type=constants.OVN_VIF_PORT_TYPES[0],
            mac=['aa:bb:cc:dd:ee:ff 10.10.1.16 2002::1234:abcd:ffff:c0a8:101'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ip.assert_called_once_with(
            ['10.10.1.16', '2002::1234:abcd:ffff:c0a8:101'], row)

    def test_run_wrong_type(self):
        row = utils.create_row(type='farofa',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ip.assert_not_called()


class TestPortBindingChassisDeletedEvent(test_base.TestCase):

    def setUp(self):
        super(TestPortBindingChassisDeletedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = bgp_watcher.PortBindingChassisDeletedEvent(self.agent)

    def test_match_fn(self):
        event = self.event.ROW_DELETE
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_not_single_or_dual_stack(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_update(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[ch])
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_update_old_chassis_set_up_false(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               up=[True])
        old = utils.create_row(chassis=[ch],
                               up=[False])
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_update_old_chassis_set_up_true(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               up=[False])
        old = utils.create_row(chassis=[ch],
                               up=[True])
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_update_different_chassis_set_up_false(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name='other-chassis')
        row = utils.create_row(chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               up=[False])
        old = utils.create_row(up=[True])
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_update_no_chassis(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[ch])
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_update_different_chassis(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name=self.chassis)
        ch_new = utils.create_row(name='new-chassis')
        row = utils.create_row(chassis=[ch_new],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[ch])
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_index_error(self):
        row = utils.create_row(chassis=[],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_run(self):
        row = utils.create_row(type=constants.OVN_VIF_PORT_TYPES[0],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_ip.assert_called_once_with(['10.10.1.16'], row)

    def test_run_dual_stack(self):
        row = utils.create_row(
            type=constants.OVN_VIF_PORT_TYPES[0],
            mac=['aa:bb:cc:dd:ee:ff 10.10.1.16 2002::1234:abcd:ffff:c0a8:101'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_ip.assert_called_once_with(
            ['10.10.1.16', '2002::1234:abcd:ffff:c0a8:101'], row)

    def test_run_wrong_type(self):
        row = utils.create_row(type='farofa',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_ip.assert_not_called()


class TestFIPSetEvent(test_base.TestCase):

    def setUp(self):
        super(TestFIPSetEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = bgp_watcher.FIPSetEvent(self.agent)

    def test_match_fn(self):
        row = utils.create_row(chassis=[], nat_addresses=['10.10.1.16'],
                               logical_port='fake-lp')
        old = utils.create_row(nat_addresses=['10.10.1.17'])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_same_nat_addresses(self):
        row = utils.create_row(chassis=[], nat_addresses=['10.10.1.16'],
                               logical_port='fake-lp')
        old = utils.create_row(nat_addresses=['10.10.1.16'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_chassis_set(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch], nat_addresses=['10.10.1.16'],
                               logical_port='fake-lp')
        old = utils.create_row(nat_addresses=['10.10.1.17'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_lrp(self):
        row = utils.create_row(chassis=[], nat_addresses=['10.10.1.16'],
                               logical_port='lrp-fake')
        old = utils.create_row(nat_addresses=['10.10.1.17'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_attribute_errir(self):
        row = utils.create_row()
        old = utils.create_row(nat_addresses=['10.10.1.17'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE,
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.17 '
                'is_chassis_resident(\\"cr-lrp-aaaaaaaa-aaaa-aaaa-'
                'aaaa-aaaaaaaaaaaa\\")'])
        old = utils.create_row(
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.18 '
                'is_chassis_resident(\\"cr-lrp-bbbbbbbb-bbbb-bbbb-'
                'bbbb-bbbbbbbbbbbb\\")'])
        self.event.run(mock.Mock(), row, old)
        self.agent.expose_ip.assert_called_once_with(
            ['10.10.1.16', '10.10.1.17'], row,
            associated_port='cr-lrp-aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa\\')

    def test_run_same_port(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE,
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.17 '
                'is_chassis_resident(\\"cr-lrp-aaaaaaaa-aaaa-aaaa-'
                'aaaa-aaaaaaaaaaaa\\")'])
        old = utils.create_row(
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.18 '
                'is_chassis_resident(\\"cr-lrp-aaaaaaaa-aaaa-aaaa-'
                'aaaa-aaaaaaaaaaaa\\")'])
        self.event.run(mock.Mock(), row, old)
        self.agent.expose_ip.assert_called_once_with(
            ['10.10.1.17'], row,
            associated_port='cr-lrp-aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa\\')

    def test_run_empty_old_nat_addresses(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE,
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.17 '
                'is_chassis_resident(\\"cr-lrp-aaaaaaaa-aaaa-aaaa-'
                'aaaa-aaaaaaaaaaaa\\")'])
        old = utils.create_row(nat_addresses=[])
        self.event.run(mock.Mock(), row, old)
        self.agent.expose_ip.assert_called_once_with(
            ['10.10.1.16', '10.10.1.17'], row,
            associated_port='cr-lrp-aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa\\')

    def test_run_wrong_type(self):
        row = utils.create_row(
            type='feijoada',
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.17 '
                'is_chassis_resident(\\"cr-lrp-aaaaaaaa-aaaa-aaaa-'
                'aaaa-aaaaaaaaaaaa\\")'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ip.assert_not_called()


class TestFIPUnsetEvent(test_base.TestCase):

    def setUp(self):
        super(TestFIPUnsetEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = bgp_watcher.FIPUnsetEvent(self.agent)

    def test_match_fn(self):
        row = utils.create_row(chassis=[], nat_addresses=['10.10.1.16'],
                               logical_port='fake-lp')
        old = utils.create_row(nat_addresses=['10.10.1.17'])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_same_nat_addresses(self):
        row = utils.create_row(chassis=[], nat_addresses=['10.10.1.16'],
                               logical_port='fake-lp')
        old = utils.create_row(nat_addresses=['10.10.1.16'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_chassis_set(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(chassis=[ch], nat_addresses=['10.10.1.16'],
                               logical_port='fake-lp')
        old = utils.create_row(nat_addresses=['10.10.1.17'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_lrp(self):
        row = utils.create_row(chassis=[], nat_addresses=['10.10.1.16'],
                               logical_port='lrp-fake')
        old = utils.create_row(nat_addresses=['10.10.1.17'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_attribute_errir(self):
        row = utils.create_row()
        old = utils.create_row(nat_addresses=['10.10.1.17'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE,
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.17 '
                'is_chassis_resident(\\"cr-lrp-aaaaaaaa-aaaa-aaaa-'
                'aaaa-aaaaaaaaaaaa\\")'])
        old = utils.create_row(
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.18 '
                'is_chassis_resident(\\"cr-lrp-bbbbbbbb-bbbb-bbbb-'
                'bbbb-bbbbbbbbbbbb\\")'])
        self.event.run(mock.Mock(), row, old)
        self.agent.withdraw_ip.assert_called_once_with(
            ['10.10.1.16', '10.10.1.18'], row,
            associated_port='cr-lrp-bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb\\')

    def test_run_same_port(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE,
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.17 '
                'is_chassis_resident(\\"cr-lrp-aaaaaaaa-aaaa-aaaa-'
                'aaaa-aaaaaaaaaaaa\\")'])
        old = utils.create_row(
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.18 '
                'is_chassis_resident(\\"cr-lrp-aaaaaaaa-aaaa-aaaa-'
                'aaaa-aaaaaaaaaaaa\\")'])
        self.event.run(mock.Mock(), row, old)
        self.agent.withdraw_ip.assert_called_once_with(
            ['10.10.1.18'], row,
            associated_port='cr-lrp-aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa\\')

    def test_run_empty_row_nat_addresses(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE,
            nat_addresses=[])
        old = utils.create_row(
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.17 '
                'is_chassis_resident(\\"cr-lrp-aaaaaaaa-aaaa-aaaa-'
                'aaaa-aaaaaaaaaaaa\\")'])
        self.event.run(mock.Mock(), row, old)
        self.agent.withdraw_ip.assert_called_once_with(
            ['10.10.1.16', '10.10.1.17'], row,
            associated_port='cr-lrp-aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa\\')

    def test_run_wrong_type(self):
        row = utils.create_row(
            type='feijoada',
            nat_addresses=[
                'aa:aa:aa:aa:aa:aa 10.10.1.16 10.10.1.17 '
                'is_chassis_resident(\\"cr-lrp-aaaaaaaa-aaaa-aaaa-'
                'aaaa-aaaaaaaaaaaa\\")'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_ip.assert_not_called()


class TestSubnetRouterAttachedEvent(test_base.TestCase):

    def setUp(self):
        super(TestSubnetRouterAttachedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = bgp_watcher.SubnetRouterAttachedEvent(self.agent)

    def test_match_fn(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_single_or_dual_stack(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff'],
                               options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_lrp(self):
        row = utils.create_row(chassis=[], logical_port='fake-lp',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_chassis_redirect(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={'chassis-redirect-port': True})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_chassis_set(self):
        row = utils.create_row(chassis=[mock.Mock()], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_index_error(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=[], options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(type=constants.OVN_PATCH_VIF_PORT_TYPE,
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_subnet.assert_called_once_with('10.10.1.16', row)

    def test_run_wrong_type(self):
        row = utils.create_row(type='coxinha',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_subnet.assert_not_called()


class TestSubnetRouterUpdateEvent(test_base.TestCase):

    def setUp(self):
        super(TestSubnetRouterUpdateEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = bgp_watcher.SubnetRouterUpdateEvent(self.agent)

    def test_match_fn(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_single_or_dual_stack(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff'],
                               options={})
        old = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff'],
                               options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_not_lrp(self):
        row = utils.create_row(chassis=[], logical_port='fake-lp',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_chassis_redirect(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={'chassis-redirect-port': True})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_chassis_set(self):
        row = utils.create_row(chassis=[mock.Mock()], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_mac_not_changed(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        old = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_mac_changed(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        old = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16 10.10.1.17'],
                               options={})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_index_error(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=[], options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(type=constants.OVN_PATCH_VIF_PORT_TYPE,
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(type=constants.OVN_PATCH_VIF_PORT_TYPE,
                               mac=['aa:bb:cc:dd:ee:ff'])
        self.event.run(mock.Mock(), row, old)
        self.agent.update_subnet.assert_called_once_with(old, row)

    def test_run_wrong_type(self):
        row = utils.create_row(type='coxinha',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(type='coxinha',
                               mac=['aa:bb:cc:dd:ee:ff'])
        self.event.run(mock.Mock(), row, old)
        self.agent.update_subnet.assert_not_called()


class TestSubnetRouterDetachedEvent(test_base.TestCase):

    def setUp(self):
        super(TestSubnetRouterDetachedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = bgp_watcher.SubnetRouterDetachedEvent(self.agent)

    def test_match_fn(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_single_or_dual_stack(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff'],
                               options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_lrp(self):
        row = utils.create_row(chassis=[], logical_port='fake-lp',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_chassis_redirect(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={'chassis-redirect-port': True})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_chassis_set(self):
        row = utils.create_row(chassis=[mock.Mock()], logical_port='lrp-fake',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_index_error(self):
        row = utils.create_row(chassis=[], logical_port='lrp-fake',
                               mac=[], options={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(type=constants.OVN_PATCH_VIF_PORT_TYPE,
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_subnet.assert_called_once_with('10.10.1.16', row)

    def test_run_wrong_type(self):
        row = utils.create_row(type='coxinha',
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_subnet.assert_not_called()


class TestTenantPortCreatedEvent(test_base.TestCase):

    def setUp(self):
        super(TestTenantPortCreatedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.agent.ovn_local_lrps = ['172.24.100.111']
        self.event = bgp_watcher.TenantPortCreatedEvent(self.agent)

    def test_match_fn(self):
        row = utils.create_row(chassis=[mock.Mock()],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_unknown_mac(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(chassis=[mock.Mock()],
                               mac=['unknown'],
                               external_ids={
                                   'neutron:cidrs': '10.10.1.16/24'})
        old = utils.create_row(chassis=[],
                               mac=['unknown'],
                               external_ids={
                                   'neutron:cidrs': '10.10.1.16/24'})
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_unknown_mac_no_cidr(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(chassis=[mock.Mock()],
                               mac=['unknown'],
                               external_ids={})
        old = utils.create_row(chassis=[],
                               mac=['unknown'],
                               external_ids={})
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_no_chassis(self):
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_not_single_or_dual_stack(self):
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_old_chassis_set(self):
        row = utils.create_row(chassis=[mock.Mock()],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[mock.Mock()])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_empty_ovn_local_lrps(self):
        self.agent.ovn_local_lrps = []
        row = utils.create_row(chassis=[mock.Mock()],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_index_error(self):
        row = utils.create_row(chassis=[mock.Mock()], mac=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_remote_ip.assert_called_once_with(
            ['10.10.1.16'], row)

    def test_run_unknown_mac(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               mac=['unknown'],
                               external_ids={
                                   'neutron:cidrs': '10.10.1.16/24'})
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_remote_ip.assert_called_once_with(
            ['10.10.1.16'], row)

    def test_run_dual_stack(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            mac=['aa:bb:cc:dd:ee:ff 10.10.1.16 2002::1234:abcd:ffff:c0a8:101'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_remote_ip.assert_called_once_with(
            ['10.10.1.16', '2002::1234:abcd:ffff:c0a8:101'], row)

    def test_run_wrong_type(self):
        row = utils.create_row(type='brigadeiro')
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_remote_ip.assert_not_called()


class TestTenantPortDeletedEvent(test_base.TestCase):

    def setUp(self):
        super(TestTenantPortDeletedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.agent.ovn_local_lrps = ['172.24.100.111']
        self.event = bgp_watcher.TenantPortDeletedEvent(self.agent)

    def test_match_fn(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(chassis=[],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[mock.Mock()],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_unknown_mac(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(chassis=[],
                               mac=['unknown'],
                               external_ids={
                                   'neutron:cidrs': '192.168.1.10/24'})
        old = utils.create_row(chassis=[mock.Mock()],
                               mac=['unknown'],
                               external_ids={
                                   'neutron:cidrs': '192.168.1.10/24'})
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_unknown_mac_no_cidr(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(chassis=[],
                               mac=['unknown'],
                               external_ids={})
        old = utils.create_row(chassis=[mock.Mock()],
                               mac=['unknown'],
                               external_ids={})
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(chassis=[],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.assertTrue(self.event.match_fn(event, row, mock.Mock()))

    def test_match_fn_not_single_or_dual_stack(self):
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_empty_ovn_local_lrps(self):
        row = utils.create_row(chassis=[],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[mock.Mock()],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.agent.ovn_local_lrps = []
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_index_error(self):
        row = utils.create_row(mac=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               chassis=[mock.Mock()])
        self.event.run(event, row, mock.Mock())
        self.agent.withdraw_remote_ip.assert_called_once_with(
            ['10.10.1.16'], row, mock.ANY)

    def test_run_unknown_mac(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               chassis=[mock.Mock()],
                               mac=['unknown'],
                               external_ids={
                                   'neutron:cidrs': '10.10.1.16/24'})
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_remote_ip.assert_called_once_with(
            ['10.10.1.16'], row, mock.ANY)

    def test_run_dual_stack(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            mac=['aa:bb:cc:dd:ee:ff 10.10.1.16 2002::1234:abcd:ffff:c0a8:101'],
            chassis=[mock.Mock()])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_remote_ip.assert_called_once_with(
            ['10.10.1.16', '2002::1234:abcd:ffff:c0a8:101'], row, mock.ANY)

    def test_run_wrong_type(self):
        row = utils.create_row(type='brigadeiro')
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_remote_ip.assert_not_called()

    def test_run_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               chassis=[mock.Mock()])
        self.event.run(event, row, [])
        self.agent.withdraw_remote_ip.assert_called_once_with(
            ['10.10.1.16'], row, mock.ANY)


class OVNLBVIPPortEvent(test_base.TestCase):

    def setUp(self):
        super(OVNLBVIPPortEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.agent.ovn_local_cr_lrps = {'fake-cr-lrp-port': {}}
        self.event = bgp_watcher.OVNLBVIPPortEvent(self.agent)

    def test_match_fn(self):
        row = utils.create_row(chassis=[], mac=[], up=[False])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_chassis(self):
        row = utils.create_row(chassis=[mock.Mock()], mac=[], up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_mac(self):
        row = utils.create_row(chassis=[],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               up=['False'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_up(self):
        row = utils.create_row(chassis=[], mac=[], up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_empty_ovn_local_cr_lrps(self):
        self.agent.ovn_local_cr_lrps = []
        row = utils.create_row(chassis=[], mac=[], up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_attribute_error(self):
        row = utils.create_row(mac=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        event = self.event.ROW_CREATE
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE, mac=[], chassis=[],
            up=[False], external_ids={
                constants.OVN_CIDRS_EXT_ID_KEY: "10.10.1.16/24"})
        self.event.run(event, row, mock.Mock())
        self.agent.expose_ovn_lb.assert_called_once_with(
            '10.10.1.16', row)

    def test_run_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE, mac=[], chassis=[],
            up=[False], external_ids={
                constants.OVN_CIDRS_EXT_ID_KEY: "10.10.1.16/24"})
        self.event.run(event, row, mock.Mock())
        self.agent.withdraw_ovn_lb.assert_called_once_with(
            '10.10.1.16', row)

    def test_run_no_external_id(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE, mac=[], chassis=[],
            up=[False], external_ids={})
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ovn_lb.assert_not_called()
        self.agent.withdraw_ovn_lb.assert_not_called()

    def test_run_wrong_type(self):
        row = utils.create_row(
            type=constants.OVN_VIRTUAL_VIF_PORT_TYPE, mac=[], chassis=[],
            up=[False], external_ids={
                constants.OVN_CIDRS_EXT_ID_KEY: "10.10.1.16/24"})
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ovn_lb.assert_not_called()
        self.agent.withdraw_ovn_lb.assert_not_called()


class TestOVNLBMemberCreateEvent(test_base.TestCase):

    def setUp(self):
        super(TestOVNLBMemberCreateEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.agent.ovn_local_cr_lrps = {
            'cr-lrp1': {'provider_datapath': 'dp1',
                        'subnets_datapath': {'lrp1': 's_dp1'},
                        'ovn_lbs': 'ovn-lb1'}}
        self.event = bgp_watcher.OVNLBMemberCreateEvent(self.agent)

    def test_match_fn(self):
        self.assertTrue(self.event.match_fn(mock.Mock(), mock.Mock(),
                                            mock.Mock()))

    def test_match_fn_no_cr_lrp(self):
        self.agent.ovn_local_cr_lrps = {}
        self.assertFalse(self.event.match_fn(mock.Mock(), mock.Mock(),
                                             mock.Mock()))

    def test_run(self):
        dpg1 = utils.create_row(_uuid='fake_dp_group',
                                datapaths=['s_dp1'])
        row = utils.create_row(name='ovn-lb1',
                               datapath_group=[dpg1],
                               vips={'172.24.100.66:80': '10.0.0.5:8080'})
        vip_port = utils.create_row(
            datapath='dp1',
            logical_port='ovn-lb-port-1',
            external_ids={constants.OVN_CIDRS_EXT_ID_KEY: '172.24.100.66/26'})
        self.agent.sb_idl.get_ovn_vip_port.return_value = vip_port
        self.agent.sb_idl.get_lrps_for_datapath.return_value = ['fake-lrp']
        self.agent.sb_idl.get_port_datapath.return_value = 'r_dp'
        self.event.run(self.event.ROW_CREATE, row, mock.Mock())
        self.agent.expose_ovn_lb_on_provider.assert_called_once_with(
            '172.24.100.66', row.name, 'cr-lrp1')
        self.agent.withdraw_ovn_lb_on_provider.assert_not_called()

    def test_run_no_subnets_datapath(self):
        CONF.set_override('expose_tenant_networks', False)
        self.addCleanup(CONF.clear_override, 'expose_tenant_networks')
        dpg1 = utils.create_row(_uuid='fake_dp_group',
                                datapaths=['s_dp1'])
        row = utils.create_row(name='ovn-lb1',
                               ls_datapath_group=[dpg1],
                               vips={'172.24.100.66:80': '10.0.0.5:8080'})
        self.agent.ovn_local_cr_lrps = {
            'cr-lrp1': {'provider_datapath': 'dp1',
                        'router_datapath': 'r_dp',
                        'subnets_datapath': {},
                        'ovn_lbs': 'ovn-lb1'}}
        vip_port = utils.create_row(
            datapath='dp1',
            logical_port='ovn-lb-port-1',
            external_ids={constants.OVN_CIDRS_EXT_ID_KEY: '172.24.100.66/26'})
        self.agent.sb_idl.get_ovn_vip_port.return_value = vip_port
        self.agent.sb_idl.get_lrps_for_datapath.return_value = ['fake-lrp']
        self.agent.sb_idl.get_port_datapath.return_value = 'r_dp'
        self.event.run(self.event.ROW_CREATE, row, mock.Mock())
        self.agent.expose_ovn_lb_on_provider.assert_called_once_with(
            '172.24.100.66', row.name, 'cr-lrp1')
        self.agent.withdraw_ovn_lb_on_provider.assert_not_called()

    def test_run_no_vip_port(self):
        dpg1 = utils.create_row(_uuid='fake_dp_group',
                                datapaths=['s_dp1'])
        row = utils.create_row(name='ovn-lb1',
                               lr_datapath_group=[dpg1],
                               vips={'172.24.100.66:80': '10.0.0.5:8080'})
        self.agent.sb_idl.get_ovn_vip_port.return_value = []
        self.event.run(self.event.ROW_CREATE, row, mock.Mock())
        self.agent.expose_ovn_lb_on_provider.assert_not_called()
        self.agent.withdraw_ovn_lb_on_provider.assert_not_called()

    def test_run_different_provider(self):
        dpg1 = utils.create_row(_uuid='fake_dp_group',
                                datapaths=['s_dp1'])
        row = utils.create_row(name='ovn-lb1',
                               datapath_group=[dpg1],
                               vips={'172.24.100.66:80': '10.0.0.5:8080'})
        vip_port = utils.create_row(
            datapath='dp2',
            logical_port='ovn-lb-port-1',
            external_ids={constants.OVN_CIDRS_EXT_ID_KEY: '172.24.100.66/26'})
        self.agent.sb_idl.get_lrps_for_datapath.return_value = ['fake-lrp']
        self.agent.sb_idl.get_port_datapath.return_value = 'dp2'
        self.agent.sb_idl.get_ovn_vip_port.return_value = vip_port
        self.event.run(self.event.ROW_CREATE, row, mock.Mock())
        self.agent.expose_ovn_lb_on_provider.assert_not_called()
        self.agent.withdraw_ovn_lb_on_provider.assert_not_called()

    def test_run_no_cr_lrp_match(self):
        dpg1 = utils.create_row(_uuid='fake_dp_group',
                                datapaths=['s_dp2'])
        row = utils.create_row(name='ovn-lb1',
                               ls_datapath_group=[dpg1],
                               vips={'172.24.100.66:80': '10.0.0.5:8080'})
        vip_port = utils.create_row(
            datapath='dp1',
            logical_port='ovn-lb-port-1',
            external_ids={constants.OVN_CIDRS_EXT_ID_KEY: '172.24.100.66/26'})
        self.agent.sb_idl.get_lrps_for_datapath.return_value = ['fake-lrp']
        self.agent.sb_idl.get_port_datapath.return_value = 'r_dp'
        self.agent.sb_idl.get_ovn_vip_port.return_value = vip_port
        self.event.run(self.event.ROW_CREATE, row, mock.Mock())
        self.agent.expose_ovn_lb_on_provider.assert_not_called()
        self.agent.withdraw_ovn_lb_on_provider.assert_not_called()

    def test_run_no_vip(self):
        dpg1 = utils.create_row(_uuid='fake_dp_group',
                                datapaths=['s_dp1'])
        row = utils.create_row(name='ovn-lb1',
                               lr_datapath_group=[dpg1],
                               vips={'172.24.100.66:80': '10.0.0.5:8080'})
        vip_port = utils.create_row(
            datapath='dp1',
            logical_port='port-1',
            external_ids={})
        self.agent.sb_idl.get_lrps_for_datapath.return_value = ['fake-lrp']
        self.agent.sb_idl.get_port_datapath.return_value = 'r_dp'
        self.agent.sb_idl.get_ovn_vip_port.return_value = vip_port
        self.event.run(self.event.ROW_CREATE, row, mock.Mock())
        self.agent.expose_ovn_lb_on_provider.assert_not_called()
        self.agent.withdraw_ovn_lb_on_provider.assert_not_called()


class TestOVNLBMemberDeleteEvent(test_base.TestCase):

    def setUp(self):
        super(TestOVNLBMemberDeleteEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.agent.provider_ovn_lbs = {
            'ovn-lb1': {'ips': ['fake-ip'], 'gateway_port': 'cr-lrp1'}}
        self.event = bgp_watcher.OVNLBMemberDeleteEvent(self.agent)

    def test_match_fn(self):
        row = utils.create_row(name='ovn-lb1')
        self.assertTrue(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_no_lb(self):
        row = utils.create_row(name='ovn-lb2')
        self.agent.ovn_local_cr_lrps = {}
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        dpg1 = utils.create_row(_uuid='fake_dp_group',
                                datapaths=['s_dp1'])
        row = utils.create_row(name='ovn-lb1',
                               datapath_group=[dpg1],
                               vips={'172.24.100.66:80': '10.0.0.5:8080'})
        vip_port = utils.create_row(
            datapath='dp1',
            logical_port='ovn-lb-port-1',
            external_ids={constants.OVN_CIDRS_EXT_ID_KEY: '172.24.100.66/26'})
        self.agent.sb_idl.get_lrps_for_datapath.return_value = ['fake-lrp']
        self.agent.sb_idl.get_port_datapath.return_value = 'r_dp'
        self.agent.sb_idl.get_ovn_vip_port.return_value = vip_port
        self.event.run(self.event.ROW_DELETE, row, mock.Mock())
        self.agent.withdraw_ovn_lb_on_provider.assert_called_once_with(
            row.name, 'cr-lrp1')
        self.agent.expose_ovn_lb_on_provider.assert_not_called()


class TestLocalnetCreateDeleteEvent(test_base.TestCase):

    def setUp(self):
        super(TestLocalnetCreateDeleteEvent, self).setUp()
        self.agent = mock.Mock()
        self.event = bgp_watcher.LocalnetCreateDeleteEvent(self.agent)

    def test_match_fn(self):
        row = utils.create_row(type=constants.OVN_LOCALNET_VIF_PORT_TYPE)
        self.assertTrue(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_match(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE)
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        self.event.run(mock.Mock(), mock.Mock(), mock.Mock())
        self.agent.sync.assert_called_once()



# Copyright 2024 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import loadbalancer as lb_utils
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests import utils as test_utils


class TestGetVipsFromLb(test_base.TestCase):
    def test_get_vips(self):
        vips = {'192.168.1.50:80': '192.168.1.100:80',
                '172.24.4.5:80': '192.168.1.100:80'}
        expected_vip_set = {"192.168.1.50", "172.24.4.5"}
        row = test_utils.create_row(vips=vips)
        observed = lb_utils.get_vips(row)

        self.assertSetEqual(expected_vip_set, observed)

    def test_get_vips_not_present(self):
        row = test_utils.create_row()
        observed = lb_utils.get_vips(row)

        self.assertSetEqual(set(), observed)


class TestIsVip(test_base.TestCase):
    def test_is_vip(self):
        ip = 'ip'
        row = test_utils.create_row(
            external_ids={constants.OVN_LB_VIP_IP_EXT_ID_KEY: ip})

        self.assertTrue(lb_utils.is_vip(row, ip))


class TestIsFip(test_base.TestCase):
    def test_is_fip(self):
        ip = 'ip'
        row = test_utils.create_row(
            external_ids={constants.OVN_LB_VIP_FIP_EXT_ID_KEY: ip})

        self.assertTrue(lb_utils.is_fip(row, ip))



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unittest import mock

from ovn_bgp_agent.tests import base as test_base


class TestAgentCmd(test_base.TestCase):
    @mock.patch('ovn_bgp_agent.agent.start')
    def test_start(self, m_start):
        from ovn_bgp_agent.cmd import agent  # To make it import a mock.
        agent.start()

        m_start.assert_called()



# Copyright 2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import ovn as ovn_utils
from ovn_bgp_agent.tests.functional import base


class OvsdbNbOvnIdl(base.BaseFunctionalNorthboundTestCase):
    def _lsp_add(self, ls_name, lsp_name, type_, tag):
        self.nb_api.lsp_add(ls_name, lsp_name, type=type_).execute(
            check_error=True)
        # lsp_add requires parent to be specified with the tag, work it
        # around with the db_set
        self.nb_api.db_set(
            'Logical_Switch_Port', lsp_name, ('tag', tag)).execute(
                check_error=True)

    def test_get_network_vlan_tags(self):
        # 0 is not a valid tag, let's start with 1
        expected_tags = list(range(1, 4))
        len_tags = len(expected_tags)

        for i, tag in enumerate(expected_tags):
            self.nb_api.ls_add('ls%d' % i).execute(check_error=True)
            ls_name = 'ls%d' % (i % 2)
            lsp_name = 'localnetport%d' % i
            self._lsp_add(
                ls_name, lsp_name,
                constants.OVN_LOCALNET_VIF_PORT_TYPE, tag=tag)
        for i, tag in enumerate(expected_tags):
            ls_name = 'ls%d' % i
            lsp_name = 'port%d' % i
            self._lsp_add(
                ls_name, lsp_name,
                type_=None, tag=i + len_tags)

        tags = self.nb_api.get_network_vlan_tags()
        self.assertCountEqual(expected_tags, tags)

    def test_get_distributed_flag_default(self):
        self.nb_api.db_remove(
            'NB_Global', '.', 'external_ids',
            constants.OVN_FIP_DISTRIBUTED, if_exists=True).execute(
                check_error=True)
        self.assertTrue(self.nb_api.get_distributed_flag())

    def test_get_distributed_flag_True(self):
        self.nb_api.db_set(
            'NB_Global', '.',
            ('external_ids', {constants.OVN_FIP_DISTRIBUTED: "True"})).execute(
                check_error=True)
        self.assertTrue(self.nb_api.get_distributed_flag())

    def test_get_distributed_flag_False(self):
        self.nb_api.db_set(
            'NB_Global', '.',
            ('external_ids', {
                constants.OVN_FIP_DISTRIBUTED: "False"})).execute(
                    check_error=True)
        self.assertFalse(self.nb_api.get_distributed_flag())

    def test_get_nats_by_lrp(self):
        router_name = 'router'
        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.lr_add(router_name, nat=[]))
            snat = txn.add(self.nb_api.lr_nat_add(
                router_name,
                constants.OVN_SNAT,
                '10.0.0.1',
                '192.168.0.1/24'))
            dnat_lrp = txn.add(self.nb_api.lr_nat_add(
                router_name,
                constants.OVN_DNAT_AND_SNAT,
                '10.0.0.10',
                '192.168.0.10'))
            dnat1_lrp = txn.add(self.nb_api.lr_nat_add(
                router_name,
                constants.OVN_DNAT_AND_SNAT,
                '10.0.0.11',
                '192.168.0.11'))
            dnat_lrp2 = txn.add(self.nb_api.lr_nat_add(
                router_name,
                constants.OVN_DNAT_AND_SNAT,
                '10.0.0.20',
                '192.168.0.20'))
            lrp1 = txn.add(self.nb_api.lrp_add(
                router_name, 'lrp', '00:00:00:00:00:01', ['10.0.0.0/24']))
            lrp2 = txn.add(self.nb_api.lrp_add(
                router_name, 'lrp2', '00:00:00:00:00:02', ['10.0.2.0/24']))

        with self.nb_api.transaction(check_error=True) as txn:
            for nat, lrp in [
                    (snat.result, lrp1.result),
                    (dnat_lrp.result, lrp1.result),
                    (dnat1_lrp.result, lrp1.result),
                    (dnat_lrp2.result, lrp2.result)]:
                txn.add(self.nb_api.db_set(
                    'NAT', nat.uuid, ('gateway_port', lrp.uuid)))

        nats = self.nb_api.get_nats_by_lrp(lrp1.result)
        self.assertCountEqual(
            [dnat_lrp.result.external_ip, dnat1_lrp.result.external_ip],
            [nat.external_ip for nat in nats])


class GetLSPsForGwChassisCommandTestCase(
        base.BaseFunctionalNorthboundTestCase):
    # format is 10.0.<router>.<switch>*10 + <port>
    FIP_TEMP = '10.0.%d.%d'
    GW_MAC_TEMP = '00:00:00:00:00:%d0'

    def setUp(self):
        super().setUp()
        with self.nb_api.transaction(check_error=True) as txn:
            for switch in range(6):
                txn.add(self.nb_api.ls_add('ls%d' % switch))
                for switch_port in range(4):
                    txn.add(self.nb_api.lsp_add(
                        'ls%d' % switch, 'lsp%d-%d' % (switch, switch_port)))
            for router in range(3):
                router_name = 'lr%d' % router
                txn.add(self.nb_api.lr_add(router_name, nat=[]))
                # gateway port
                txn.add(self.nb_api.lrp_add(
                    'lr%d' % router,
                    'lrp%d-0' % router,
                    self.GW_MAC_TEMP % router,
                    ['10.0.%d.0/24' % router],
                    status={
                        constants.OVN_STATUS_CHASSIS: 'chassis%d' % (
                            router % 2)}))
                # connect two switches to the router
                for lrp_index, lrp_peer in enumerate([1, 2]):
                    ls_index = router + lrp_peer * 2
                    network = '192.168.%d.0/24' % ls_index
                    txn.add(self.nb_api.lrp_add(
                        router_name,
                        'lrp%d-%d' % (router, lrp_peer),
                        '00:00:00:00:00:%d%d' % (router, lrp_index),
                        [network],
                        peer='lsp%d-0' % ls_index))
                    txn.add(self.nb_api.lr_nat_add(
                        router_name,
                        constants.OVN_SNAT,
                        '10.0.%d.1' % router,
                        network))

            # Add DNAT rules to ports on one switch 2 NATs, then 1 NAT,
            # and so on
            for router, switch in enumerate(range(0, 6, 2)):
                for lsp in range(1, 3):
                    txn.add(self.nb_api.lr_nat_add(
                        'lr%d' % router,
                        constants.OVN_DNAT_AND_SNAT,
                        self.FIP_TEMP % (router, switch * 10 + lsp),
                        '192.168.%d.%d' % (switch, lsp),
                        logical_port='lsp%d-%d' % (switch, lsp),
                        external_mac='00:00:00:00:0%d:0%d' % (switch, lsp)))
                switch += 1
                txn.add(self.nb_api.lr_nat_add(
                    'lr%d' % router,
                    constants.OVN_DNAT_AND_SNAT,
                    self.FIP_TEMP % (router, switch * 10 + 1),
                    '192.168.%d.1' % switch,
                    logical_port='lsp%d-1' % switch,
                    external_mac='00:00:00:00:0%d:01' % switch))

        # Set the gateway ports for all NATs
        for router in range(3):
            for nat in self.nb_api.lr_nat_list(
                    'lr%d' % router).execute(check_error=True):
                lrp = self.nb_api.lrp_get(
                    'lrp%d-0' % router).execute(check_error=True)
                self.nb_api.db_set(
                    'NAT',
                    nat.uuid,
                    ('gateway_port', lrp),
                    ('external_ids', {
                        constants.OVN_FIP_NET_EXT_ID_KEY: 'external'})
                ).execute(check_error=True)

    def _make_fip_result(self, router_index, switch_index, lsp_index):
        lsp = self.nb_api.lsp_get(
            'lsp%d-%d' % (switch_index, lsp_index)).execute(check_error=True)
        return (self.FIP_TEMP % (router_index, switch_index * 10 + lsp_index),
                self.GW_MAC_TEMP % router_index,
                'neutron-external',
                lsp)

    def _get_expected_fips(self):
        expected_found_fips = []

        for router in [0, 2]:
            switch = router * 2
            # 2 FIPs on the first switch
            for lsp in [1, 2]:
                expected_found_fips.append(
                    self._make_fip_result(router, switch, lsp))

            # 1 FIP on the second switch
            switch = router * 2 + 1
            expected_found_fips.append(
                self._make_fip_result(router, switch, lsp_index=1))

        return expected_found_fips

    def test_get_lsps(self):
        # We expect only LSPs from switches attached to routers 0 and 2
        # each router has w switches. The first switch has 2 FIPs and the
        # second has one FIP. That is 6 FIPs in total.
        expected_found_fips = self._get_expected_fips()
        result = ovn_utils.GetLSPsForGwChassisCommand(
            self.nb_api, 'chassis0').execute()

        self.assertCountEqual(expected_found_fips, result)



# Copyright 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import ipaddress
import threading

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers import driver_api
from ovn_bgp_agent.drivers.openstack import nb_exceptions
from ovn_bgp_agent.drivers.openstack.utils import bgp as bgp_utils
from ovn_bgp_agent.drivers.openstack.utils import driver_utils
from ovn_bgp_agent.drivers.openstack.utils import nat as nat_utils
from ovn_bgp_agent.drivers.openstack.utils import ovn
from ovn_bgp_agent.drivers.openstack.utils import ovs
from ovn_bgp_agent.drivers.openstack.utils import port as port_utils
from ovn_bgp_agent.drivers.openstack.utils import wire as wire_utils
from ovn_bgp_agent.drivers.openstack.watchers import nb_bgp_watcher as watcher
from ovn_bgp_agent import exceptions
from ovn_bgp_agent.utils import linux_net


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
# LOG.setLevel(logging.DEBUG)
# logging.basicConfig(level=logging.DEBUG)

OVN_TABLES = ['Logical_Switch_Port', 'NAT', 'Logical_Switch', 'Logical_Router',
              'Logical_Router_Port', 'Load_Balancer', 'DHCP_Options',
              'NB_Global']
LOCAL_CLUSTER_OVN_TABLES = ['Logical_Switch', 'Logical_Switch_Port',
                            'Logical_Router', 'Logical_Router_Port',
                            'Logical_Router_Policy',
                            'Logical_Router_Static_Route', 'Gateway_Chassis',
                            'Static_MAC_Binding']


def _validate_ovn_version(distributed, idl):
    if not distributed and 'gateway_port' not in idl.tables['NAT'].columns:
        raise RuntimeError(
            "Centralized routing requires gateway_port column in the "
            "OVN_Northbound schema. Please update OVN to 23.09.0 or later.")


class NATExposer:
    def __init__(self, agent):
        self.agent = agent

    def expose_fip_from_nat(self, nat):
        raise RuntimeError(
            "The exposer does not have distributed flag set yet")

    def withdraw_fip_from_nat(self, nat):
        raise RuntimeError(
            "The exposer does not have distributed flag set yet")

    @property
    def distributed(self):
        pass

    @distributed.setter
    def distributed(self, value):
        if value:
            self.expose_fip_from_nat = self._expose_nat_distributed
            self.withdraw_fip_from_nat = self._withdraw_nat_distributed
        else:
            self.expose_fip_from_nat = self._expose_nat_centralized
            self.withdraw_fip_from_nat = self._withdraw_nat_centralized

    def _expose_nat_distributed(self, nat):
        raise NotImplementedError("Distributed NAT is not implemented yet.")

    def _expose_nat_centralized(self, nat):
        net_id = nat.external_ids[constants.OVN_FIP_NET_EXT_ID_KEY]
        ls_name = "neutron-{}".format(net_id)

        try:
            mac = nat_utils.get_gateway_lrp(nat).mac
        except IndexError:
            LOG.error("Gateway port for NAT entry %s has no MAC address set",
                      nat.uuid)
            return

        # nat has the logical port and its in the db, that was checked in
        # match_fn
        lsp = self.agent.nb_idl.lsp_get(
            nat.logical_port[0]).execute(check_error=True)

        self.agent.expose_fip(nat.external_ip, mac, ls_name, lsp)

    def _withdraw_nat_distributed(self, nat):
        raise NotImplementedError("Distributed NAT is not implemented yet.")

    def _withdraw_nat_centralized(self, nat):
        lsp = self.agent.nb_idl.lsp_get(nat.logical_port[0]).execute()
        self.agent.withdraw_fip(nat.external_ip, lsp)


class NBOVNBGPDriver(driver_api.AgentDriverBase):

    def __init__(self):
        self.allowed_address_scopes = set(CONF.address_scopes or [])

        self._init_vars()

        self._nb_idl = None
        self._local_nb_idl = None
        self._post_start_event = threading.Event()
        self.nat_exposer = NATExposer(self)

        self.__d_events = {
            True: [
                watcher.NATMACAddedEvent(self),
                watcher.LogicalSwitchPortFIPCreateEvent(self),
                watcher.LogicalSwitchPortFIPDeleteEvent(self),
            ],
            False: [
                watcher.ExposeFIPOnCRLRP(self),
                watcher.WithdrawFIPOnCRLRP(self),
                watcher.CrLrpChassisChangeExposeEvent(self),
                watcher.CrLrpChassisChangeWithdrawEvent(self),
            ]}

    @property
    def _expose_tenant_networks(self):
        return (CONF.expose_tenant_networks or
                CONF.expose_ipv6_gua_tenant_networks)

    @property
    def nb_idl(self) -> ovn.OvsdbNbOvnIdl:
        if not self._nb_idl:
            self._post_start_event.wait()
        return self._nb_idl

    @property
    def local_nb_idl(self):
        if not self._local_nb_idl:
            self._post_start_event.wait()
        return self._local_nb_idl

    @nb_idl.setter
    def nb_idl(self, val):
        self._nb_idl = val

    @local_nb_idl.setter
    def local_nb_idl(self, val):
        self._local_nb_idl = val

    @property
    def distributed(self):
        if not hasattr(self, '_distributed'):
            self._distributed = self.nb_idl.get_distributed_flag()
            _validate_ovn_version(self._distributed, self.nb_idl.idl)
            self.nat_exposer.distributed = self._distributed
        return self._distributed

    @distributed.setter
    def distributed(self, value):
        _validate_ovn_version(value, self.nb_idl.idl)
        self._distributed = value
        self.nat_exposer.distributed = value
        self._switch_distributed_events(value)

    def _switch_distributed_events(self, distributed):
        to_unwatch = self._get_additional_events(not distributed)
        to_watch = self._get_additional_events(distributed)

        self.nb_idl.ovsdb_connection.idl.notify_handler.unwatch_events(
            to_unwatch)
        self.nb_idl.ovsdb_connection.idl.notify_handler.watch_events(
            to_watch)

    def _init_vars(self):
        self.ovn_bridge_mappings = {}  # {'public': 'br-ex'}
        self.ovs_flows = {}

        self.ovn_routing_tables = {}  # {'br-ex': 200}
        # {'br-ex': [route1, route2]}
        self.ovn_routing_tables_routes = collections.defaultdict(list)

        self.ovn_local_cr_lrps = {}
        self.ovn_local_lrps = {}

        # {'ls_name': ['ip': {'bridge_device': X, 'bridge_vlan': Y}]}
        self._exposed_ips = {}
        self._ovs_flows = collections.defaultdict()
        self.ovn_provider_ls = {}
        # dict instead of list to speed up look ups
        self.ovn_tenant_ls = {}  # {'ls_name': True}

    def start(self):
        self.ovs_idl = ovs.OvsIdl()
        self.ovs_idl.start(CONF.ovsdb_connection)
        self.chassis = self.ovs_idl.get_own_chassis_name()
        self.chassis_id = self.ovs_idl.get_own_chassis_id()

        # NOTE(ltomasbo): remote should point to NB DB port instead of SB DB,
        # so changing 6642 by 6641
        self.ovn_remote = self.ovs_idl.get_ovn_remote(nb=True)
        LOG.info("Loaded chassis %s.", self.chassis)

        LOG.info("Starting VRF configuration for advertising routes")
        # Base BGP configuration
        bgp_utils.ensure_base_bgp_configuration()

        # Clear vrf routing table
        if CONF.clear_vrf_routes_on_startup:
            linux_net.delete_routes_from_table(CONF.bgp_vrf_table_id)

        LOG.info("VRF configuration for advertising routes completed")
        if self._expose_tenant_networks and self.allowed_address_scopes:
            LOG.info("Configured allowed address scopes: %s",
                     ", ".join(self.allowed_address_scopes))

        self._post_start_event.clear()

        events = self._get_base_events()
        self.nb_idl = ovn.OvnNbIdl(
            self.ovn_remote,
            tables=OVN_TABLES,
            events=events).start()

        # if local OVN cluster, gets an idl for it
        if CONF.exposing_method == constants.EXPOSE_METHOD_OVN:
            self.local_nb_idl = ovn.OvnNbIdl(
                CONF.local_ovn_cluster.ovn_nb_connection,
                tables=LOCAL_CLUSTER_OVN_TABLES,
                events=[],
                leader_only=True).start()

        self.nb_idl.ovsdb_connection.idl.notify_handler.watch_events(
            self._get_additional_events(self.distributed))

        # Now IDL connections can be safely used
        self._post_start_event.set()

    def _get_base_events(self):
        events = {watcher.LogicalSwitchPortProviderCreateEvent(self),
                  watcher.LogicalSwitchPortProviderDeleteEvent(self),
                  watcher.OVNLBCreateEvent(self),
                  watcher.OVNLBDeleteEvent(self),
                  watcher.OVNPFCreateEvent(self),
                  watcher.OVNPFDeleteEvent(self),
                  watcher.ChassisRedirectCreateEvent(self),
                  watcher.ChassisRedirectDeleteEvent(self),
                  watcher.DistributedFlagChangedEvent(self),
                  }

        if CONF.exposing_method == constants.EXPOSE_METHOD_VRF:
            # For vrf we require more information on the logical_switch
            # before performing a sync.
            events.add(watcher.LogicalSwitchUpdateEvent(self))
        else:
            events.add(watcher.LocalnetCreateDeleteEvent(self))

        if self._expose_tenant_networks:
            events.update({watcher.LogicalSwitchPortSubnetAttachEvent(self),
                           watcher.LogicalSwitchPortSubnetDetachEvent(self)})
            if CONF.advertisement_method_tenant_networks == 'host':
                events.update({
                    watcher.LogicalSwitchPortTenantCreateEvent(self),
                    watcher.LogicalSwitchPortTenantDeleteEvent(self)
                })
        return events

    def _get_additional_events(self, distributed):
        return self.__d_events[distributed]

    @lockutils.synchronized('nbbgp')
    def frr_sync(self):
        LOG.debug("Ensuring VRF configuration for advertising routes")
        # Base BGP configuration
        bgp_utils.ensure_base_bgp_configuration()

    @lockutils.synchronized('nbbgp')
    def sync(self):
        self._init_vars()

        LOG.debug("Configuring default wiring for each provider network")
        # Apply base configuration for each bridge
        self.ovn_bridge_mappings, self.ovs_flows = (
            wire_utils.ensure_base_wiring_config(
                self.nb_idl, self.ovs_idl, ovn_idl=self.local_nb_idl,
                routing_tables=self.ovn_routing_tables))

        LOG.debug("Syncing current routes.")
        # add missing routes/ips for OVN router gateway ports
        ports = self.nb_idl.get_active_cr_lrp_on_chassis(self.chassis_id)
        for port in ports:
            self._ensure_crlrp_exposed(port)
        # add missing routes/ips for subnets connected to local gateway ports
        ports = self.nb_idl.get_active_local_lrps(
            self.ovn_local_cr_lrps.keys())
        for port in ports:
            ips = port.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                        "").split()
            subnet_info = {
                'associated_router': port.external_ids.get(
                    constants.OVN_DEVICE_ID_EXT_ID_KEY),
                'network': port.external_ids.get(
                    constants.OVN_LS_NAME_EXT_ID_KEY),
                'address_scopes': driver_utils.get_addr_scopes(port)}
            self._expose_subnet(ips, subnet_info)

        # add missing routes/ips for IPs on provider network and FIPs
        ports = self.nb_idl.get_active_lsp_on_chassis(self.chassis)
        if not self.distributed:
            # expose all FIPs if this chassis hosts the gateway port
            lsp_with_fips = ovn.GetLSPsForGwChassisCommand(
                self.nb_idl, self.chassis_id).execute(check_error=True)
            for lsp_data in lsp_with_fips:
                self._expose_fip(*lsp_data)
        for port in ports:
            if port.type not in [constants.OVN_VM_VIF_PORT_TYPE,
                                 constants.OVN_VIRTUAL_VIF_PORT_TYPE]:
                continue
            self._ensure_lsp_exposed(port)

        # add missing routes/ips for OVN loadbalancers
        self._expose_lbs(self.ovn_local_cr_lrps.keys())

        # remove extra wiring leftovers
        wire_utils.cleanup_wiring(self.nb_idl,
                                  self.ovn_bridge_mappings,
                                  self.ovs_flows,
                                  self._exposed_ips,
                                  self.ovn_routing_tables,
                                  self.ovn_routing_tables_routes)

    def _ensure_lsp_exposed(self, port):
        port_fip = port.external_ids.get(constants.OVN_FIP_EXT_ID_KEY)
        if port_fip:
            try:
                external_ip, external_mac, ls_name = (
                    self.get_port_external_ip_and_ls(port.name))
            except nb_exceptions.NATNotFound as e:
                LOG.debug("Logical Switch Port %s does not have all data "
                          "required in its NAT entry: %s", port.name, e)
                return
            return self._expose_fip(external_ip, external_mac, ls_name, port)

        logical_switch = port.external_ids.get(
            constants.OVN_LS_NAME_EXT_ID_KEY)
        if not self.is_ls_provider(logical_switch):
            return

        _, bridge_device, bridge_vlan = self._get_provider_ls_info(
            logical_switch)

        try:
            ips = port_utils.get_ips_from_lsp(port)
        except exceptions.IpAddressNotFound:
            ips = []
        mac = port_utils.get_mac_from_lsp(port)
        self._expose_ip(ips, mac, logical_switch, bridge_device, bridge_vlan,
                        port.type, port.external_ids.get(
                            constants.OVN_CIDRS_EXT_ID_KEY, "").split())

    def _ensure_crlrp_exposed(self, port):
        if not port.networks:
            return

        logical_switch = port.external_ids.get(
            constants.OVN_LS_NAME_EXT_ID_KEY)
        if not self.is_ls_provider(logical_switch):
            return

        _, bridge_device, bridge_vlan = self._get_provider_ls_info(
            logical_switch)

        ips = [net.split("/")[0] for net in port.networks]
        router = port.external_ids.get(constants.OVN_LR_NAME_EXT_ID_KEY)
        self._expose_ip(ips, port.mac, logical_switch, bridge_device,
                        bridge_vlan, constants.OVN_CR_LRP_PORT_TYPE,
                        port.networks, router=router)

    def _expose_provider_port(self, port_ips, mac, logical_switch,
                              bridge_device, bridge_vlan, localnet,
                              proxy_cidrs=None):
        if proxy_cidrs is None:
            proxy_cidrs = []
        # Connect to OVN
        try:
            if wire_utils.wire_provider_port(
                    self.ovn_routing_tables_routes, self.ovs_flows, port_ips,
                    bridge_device, bridge_vlan, localnet,
                    self.ovn_routing_tables, proxy_cidrs, mac=mac,
                    ovn_idl=self.local_nb_idl):
                # Expose the IP now that it is connected
                bgp_utils.announce_ips(port_ips)
                for ip in port_ips:
                    self._exposed_ips.setdefault(logical_switch, {}).update(
                        {ip: {'bridge_device': bridge_device,
                              'bridge_vlan': bridge_vlan}})
            else:
                return False
        except Exception as e:
            LOG.exception("Unexpected exception while wiring provider port: "
                          "%s", e)
            return False
        return True

    def _withdraw_provider_port(self, port_ips, logical_switch, bridge_device,
                                bridge_vlan, proxy_cidrs=None):
        if proxy_cidrs is None:
            proxy_cidrs = []
        # Withdraw IP before disconnecting it
        bgp_utils.withdraw_ips(port_ips)

        # Disconnect IP from OVN
        try:
            wire_utils.unwire_provider_port(
                self.ovn_routing_tables_routes, port_ips, bridge_device,
                bridge_vlan, self.ovn_routing_tables, proxy_cidrs,
                ovn_idl=self.local_nb_idl)
        except Exception as e:
            LOG.exception("Unexpected exception while unwiring provider port: "
                          "%s", e)
        for ip in port_ips:
            if self._exposed_ips.get(logical_switch, {}).get(ip):
                self._exposed_ips[logical_switch].pop(ip)

    def _get_bridge_for_localnet_port(self, localnet):
        bridge_device = None
        bridge_vlan = None
        network_name = localnet.options.get('network_name')
        if network_name:
            bridge_device = self.ovn_bridge_mappings.get(network_name)
        if localnet.tag:
            bridge_vlan = localnet.tag[0]
        return bridge_device, bridge_vlan

    def _expose_lbs(self, router_list):
        lbs = self.nb_idl.get_active_local_lbs(router_list)
        for lb in lbs:
            if driver_utils.is_pf_lb(lb):
                self._expose_ovn_pf_lb_fip(lb)
            else:
                self._expose_ovn_lb_vip(lb)
                # if vip-fip expose fip too
                if lb.external_ids.get(constants.OVN_LB_VIP_FIP_EXT_ID_KEY):
                    self._expose_ovn_lb_fip(lb)

    def _withdraw_lbs(self, router_list):
        lbs = self.nb_idl.get_active_local_lbs(router_list)
        for lb in lbs:
            if driver_utils.is_pf_lb(lb):
                self._withdraw_ovn_pf_lb_fip(lb)
            else:
                self._withdraw_ovn_lb_vip(lb)
                # if vip-fip withdraw fip too
                if lb.external_ids.get(constants.OVN_LB_VIP_FIP_EXT_ID_KEY):
                    self._withdraw_ovn_lb_fip(lb)

    def is_ls_provider(self, logical_switch):
        '''Check if given logical switch is a provider network on this host

        It will also validate that the provider network actually has been
        exposed by the driver.
        '''

        if logical_switch is None:
            self.ovn_tenant_ls[logical_switch] = True  # just for caching
            return False

        # Check if the ls has already been identified as a tenant network
        if self.ovn_tenant_ls.get(logical_switch, None) is True:
            return False

        # Check the bridge device from provider network
        _, bridge_device, _ = self._get_provider_ls_info(logical_switch)

        # Check if the bridge device has been exposed by the wiring methods
        if bridge_device not in self.ovn_bridge_mappings.values():
            return False

        return bridge_device is not None

    def is_ip_exposed(self, logical_switch, ips):
        '''Check if the ip(s) from given logical_switch is exported.

        So basically, check if the ips are listed in self._exposed_ips.
        If it is in there, it should be exposed by ovn-bgp-agent.

        This helps a lot in evaluating events.
        '''
        # Ip may be a list
        if not isinstance(ips, (list, tuple, set)):
            ips = [ips]

        for ip in ips:
            if ip in self._exposed_ips.get(logical_switch, {}):
                return True

        return False

    @lockutils.synchronized('nbbgp')
    def expose_ip(self, ips, ips_info):
        '''Advertice BGP route by adding IP to device.

        This methods ensures BGP advertises the IP of the VM in the provider
        network.
        It relies on Zebra, which creates and advertises a route when an IP
        is added to a local interface.

        This method assumes a device named self.ovn_device exists (inside a
        VRF), and adds the IP of:
        - VM IP on the provider network
        '''
        logical_switch = ips_info.get('logical_switch')
        if not self.is_ls_provider(logical_switch):
            return False

        _, bridge_device, bridge_vlan = self._get_provider_ls_info(
            logical_switch)

        mac = ips_info.get('mac')
        return self._expose_ip(ips, mac, logical_switch, bridge_device,
                               bridge_vlan, port_type=ips_info['type'],
                               cidrs=ips_info['cidrs'],
                               router=ips_info.get('router'))

    def _expose_ip(self, ips, mac, logical_switch, bridge_device, bridge_vlan,
                   port_type, cidrs, router=None):
        LOG.debug("Adding BGP route for logical port with ip %s", ips)
        localnet = self.ovn_provider_ls[logical_switch]['localnet']

        if not self._expose_provider_port(ips, mac, logical_switch,
                                          bridge_device, bridge_vlan,
                                          localnet, cidrs):
            return []

        if router and port_type == constants.OVN_CR_LRP_PORT_TYPE:
            # Store information about local CR-LRPs that will later be used
            # to expose networks
            self.ovn_local_cr_lrps[router] = {
                'bridge_device': bridge_device,
                'bridge_vlan': bridge_vlan,
                'provider_switch': logical_switch,
                'ips': ips,
            }
            # Expose associated subnets
            ports = self.nb_idl.get_active_local_lrps([router])
            for port in ports:
                ips = port.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                            "").split()
                subnet_info = {
                    'associated_router': port.external_ids.get(
                        constants.OVN_DEVICE_ID_EXT_ID_KEY),
                    'network': port.external_ids.get(
                        constants.OVN_LS_NAME_EXT_ID_KEY),
                    'address_scopes': driver_utils.get_addr_scopes(port)}
                self._expose_subnet(ips, subnet_info)

            # add missing routes/ips for OVN loadbalancers
            self._expose_lbs([router])

        LOG.debug("Added BGP route for logical port with ip %s", ips)
        return ips

    @lockutils.synchronized('nbbgp')
    def withdraw_ip(self, ips, ips_info):
        '''Withdraw BGP route by removing IP from device.

        This methods ensures BGP withdraw an advertised IP of a VM, either
        in the provider network.
        It relies on Zebra, which withdraws the advertisement as soon as the
        IP is deleted from the local interface.

        This method assumes a device named self.ovn_decice exists (inside a
        VRF), and removes the IP of:
        - VM IP on the provider network
        '''
        logical_switch = ips_info.get('logical_switch')
        if not logical_switch:
            return
        _, bridge_device, bridge_vlan = self._get_ls_localnet_info(
            logical_switch)
        if not bridge_device:
            # This means it is not a provider network
            return

        proxy_cidr = []
        if ips_info['cidrs']:
            if not (self.nb_idl.ls_has_virtual_ports(logical_switch) or
                    self.nb_idl.get_active_lsp_on_chassis(self.chassis)):
                for n_cidr in ips_info['cidrs']:
                    if (linux_net.get_ip_version(n_cidr) ==
                            constants.IP_VERSION_6):
                        proxy_cidr.append(n_cidr)
        LOG.debug("Deleting BGP route for logical port with ip %s", ips)
        self._withdraw_provider_port(ips, logical_switch, bridge_device,
                                     bridge_vlan, proxy_cidr)

        if ips_info.get('router'):
            # It is a Logical Router Port (CR-LRP)
            # Withdraw associated subnets
            ports = self.nb_idl.get_active_local_lrps([ips_info['router']])
            for port in ports:
                ips = port.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                            "").split()
                subnet_info = {
                    'associated_router': port.external_ids.get(
                        constants.OVN_DEVICE_ID_EXT_ID_KEY),
                    'network': port.external_ids.get(
                        constants.OVN_LS_NAME_EXT_ID_KEY),
                    'address_scopes': driver_utils.get_addr_scopes(port)}
                self._withdraw_subnet(ips, subnet_info)

            # withdraw routes/ips for OVN loadbalancers
            self._withdraw_lbs([ips_info['router']])

            try:
                del self.ovn_local_cr_lrps[ips_info['router']]
            except KeyError:
                LOG.debug("Gateway port for router %s already cleanup.",
                          ips_info['router'])
        LOG.debug("Deleted BGP route for logical port with ip %s", ips)

    def _get_provider_ls_info(self, logical_switch):
        '''Helper method for _get_ls_localnet_info

        It returns the information from cached self.ovn_provider_ls dictionary
        or calls the method and populates the dictionary for given
        logical_switch.

        It returns a tuple of localnet, bridge_device and bridge_vlan for
        compatibilty with _get_ls_localnet_info
        '''
        if logical_switch is None:
            return None, None, None

        if logical_switch not in self.ovn_provider_ls:
            localnet, bridge_dev, bridge_vlan = self._get_ls_localnet_info(
                logical_switch)

            self.ovn_provider_ls[logical_switch] = {
                'bridge_device': bridge_dev,
                'bridge_vlan': bridge_vlan,
                'localnet': localnet
            }

        ls = self.ovn_provider_ls[logical_switch]
        return ls['localnet'], ls['bridge_device'], ls['bridge_vlan']

    def _get_ls_localnet_info(self, logical_switch):
        localnet_ports = self.nb_idl.ls_get_localnet_ports(
            logical_switch, if_exists=True).execute(check_error=True)
        if not localnet_ports:
            # means it is not a provider network, so no need to expose the IP
            return None, None, None
        bridge_device, bridge_vlan = self._get_bridge_for_localnet_port(
            localnet_ports[0])
        # NOTE: assuming only one localnet per LS exists
        return localnet_ports[0].name, bridge_device, bridge_vlan

    def get_port_external_ip_and_ls(self, port):
        nat_entry = self.nb_idl.get_nat_by_logical_port(port)
        try:
            net_id = nat_entry.external_ids[constants.OVN_FIP_NET_EXT_ID_KEY]
        except AttributeError:
            # nat_entry is None and has no external_ids attribute
            raise nb_exceptions.NATNotFound(
                "NAT entry for port %s not found" % port)
        except KeyError:
            # network ID was not found in the NAT entry
            raise nb_exceptions.NATNotFound(
                "NAT entry for port %s does not contain network ID in its "
                "external_ids" % port)
        try:
            external_mac = nat_entry.external_mac[0]
        except IndexError:
            raise nb_exceptions.NATNotFound(
                "NAT entry for port %s does not have external_mac set" % (
                    port))

        ls_name = "neutron-{}".format(net_id)
        return nat_entry.external_ip, external_mac, ls_name

    @lockutils.synchronized('nbbgp')
    def expose_fip(self, ip, mac, logical_switch, row):
        '''Advertice BGP route by adding IP to device.

        This methods ensures BGP advertises the FIP associated to a VM in a
        tenant networks.
        It relies on Zebra, which creates and advertises a route when an IP
        is added to a local interface.

        This method assumes a device named self.ovn_device exists (inside a
        VRF), and adds the IP of:
        - VM FIP
        '''
        return self._expose_fip(ip, mac, logical_switch, row)

    def _expose_fip(self, ip, mac, logical_switch, row):
        if not self.is_ls_provider(logical_switch):
            return False

        localnet, bridge_device, bridge_vlan = self._get_provider_ls_info(
            logical_switch)

        tenant_logical_switch = row.external_ids.get(
            constants.OVN_LS_NAME_EXT_ID_KEY)
        if not tenant_logical_switch:
            return
        self.ovn_tenant_ls[tenant_logical_switch] = True
        LOG.debug("Adding BGP route for FIP with ip %s", ip)
        if not self._expose_provider_port([ip], mac, tenant_logical_switch,
                                          bridge_device, bridge_vlan,
                                          localnet):
            return False
        LOG.debug("Added BGP route for FIP with ip %s", ip)
        return True

    @lockutils.synchronized('nbbgp')
    def withdraw_fip(self, ip, row):
        '''Withdraw BGP route by removing IP from device.

        This methods ensures BGP withdraw an advertised the FIP associated to
        a VM in a tenant networks.
        It relies on Zebra, which withdraws the advertisement as soon as the
        IP is deleted from the local interface.

        This method assumes a device named self.ovn_decice exists (inside a
        VRF), and removes the IP of:
        - VM FIP
        '''
        tenant_logical_switch = row.external_ids.get(
            constants.OVN_LS_NAME_EXT_ID_KEY)
        if not tenant_logical_switch:
            return
        fip_info = self._exposed_ips.get(tenant_logical_switch, {}).get(ip)
        if not fip_info:
            # No information to withdraw the FIP
            return
        bridge_device = fip_info['bridge_device']
        bridge_vlan = fip_info['bridge_vlan']

        LOG.debug("Deleting BGP route for FIP with ip %s", ip)
        self._withdraw_provider_port([ip], tenant_logical_switch,
                                     bridge_device, bridge_vlan)
        LOG.debug("Deleted BGP route for FIP with ip %s", ip)

    @lockutils.synchronized('nbbgp')
    def expose_remote_ip(self, ips, ips_info):
        self._expose_remote_ip(ips, ips_info)

    @lockutils.synchronized('nbbgp')
    def withdraw_remote_ip(self, ips, ips_info):
        self._withdraw_remote_ip(ips, ips_info)

    def _get_exposed_ip(self, exposed_ip):
        for ls, ip_info in self._exposed_ips.items():
            if exposed_ip in ip_info:
                return ls, ip_info[exposed_ip]

    def _get_router_port_info_for_ls(self, ls):
        # LOG.debug('Searching router port info for ls %s', ls)
        lrps = list(self.ovn_local_lrps.get(ls, []))
        if not lrps:
            LOG.debug('Could not find router gateway port for ls %s', ls)
            return

        _, cr_lrp_info = self._get_exposed_ip(lrps[0])
        return cr_lrp_info

    def _expose_remote_ip(self, ips, ips_info):
        if (CONF.advertisement_method_tenant_networks ==
                constants.ADVERTISEMENT_METHOD_SUBNET):
            # Ip should already be exported via cr-lrp subnet announcement.
            return

        ips_to_expose = ips
        if not CONF.expose_tenant_networks:
            # This means CONF.expose_ipv6_gua_tenant_networks is enabled
            gua_ips = [ip for ip in ips if driver_utils.is_ipv6_gua(ip)]
            if not gua_ips:
                return
            ips_to_expose = gua_ips

        LOG.debug("Adding BGP route for tenant IP(s) %s on chassis %s",
                  ips_to_expose, self.chassis)
        if CONF.exposing_method == constants.EXPOSE_METHOD_VRF:
            cr_lrp_info = self._get_router_port_info_for_ls(
                ips_info['logical_switch'])
            if not cr_lrp_info:
                LOG.debug('Unable to export routed ips %s: could not find '
                          'router gateway port for ls %s',
                          ips, ips_info['logical_switch'])
                return

            # Add the bridge_vlan, bridge_device and via to the ips_info
            ips_info.update(cr_lrp_info)

        bgp_utils.announce_ips(ips_to_expose, ips_info=ips_info)
        for ip in ips_to_expose:
            self._exposed_ips.setdefault(
                ips_info['logical_switch'], {}).setdefault(ip, {})

        LOG.debug("Added BGP route for tenant IP(s) %s on chassis %s",
                  ips_to_expose, self.chassis)

    def _withdraw_remote_ip(self, ips, ips_info):
        if (CONF.advertisement_method_tenant_networks ==
                constants.ADVERTISEMENT_METHOD_SUBNET):
            return

        ips_to_withdraw = ips
        if not CONF.expose_tenant_networks:
            # This means CONF.expose_ipv6_gua_tenant_networks is enabled
            gua_ips = [ip for ip in ips if driver_utils.is_ipv6_gua(ip)]
            if not gua_ips:
                return
            ips_to_withdraw = gua_ips

        LOG.debug("Deleting BGP route for tenant IP(s) %s on chassis %s",
                  ips_to_withdraw, self.chassis)
        if CONF.exposing_method == constants.EXPOSE_METHOD_VRF:
            cr_lrp_info = self._get_router_port_info_for_ls(
                ips_info['logical_switch'])
            if not cr_lrp_info:
                LOG.debug('Unable to withdraw routed ips %s: could not find '
                          'router gateway port for ls %s',
                          ips, ips_info['logical_switch'])
                return

            # Add the bridge_vlan, bridge_device and via to the ips_info
            ips_info.update(cr_lrp_info)

        bgp_utils.withdraw_ips(ips_to_withdraw, ips_info=ips_info)
        for ip in ips_to_withdraw:
            self._exposed_ips.get(ips_info['logical_switch'], {}).pop(ip, None)

        LOG.debug("Deleted BGP route for tenant IP(s) %s on chassis %s",
                  ips_to_withdraw, self.chassis)

    @lockutils.synchronized('nbbgp')
    def expose_subnet(self, ips, subnet_info):
        return self._expose_subnet(ips, subnet_info)

    @lockutils.synchronized('nbbgp')
    def withdraw_subnet(self, ips, subnet_info):
        return self._withdraw_subnet(ips, subnet_info)

    def _expose_subnet(self, ips, subnet_info):
        gateway_router = subnet_info['associated_router']
        if not gateway_router:
            LOG.debug("Subnet CIDRs %s not exposed as there is no associated "
                      "router", ips)
            return
        cr_lrp_info = self.ovn_local_cr_lrps.get(gateway_router)
        if not cr_lrp_info:
            LOG.debug("Subnet CIDRs %s not exposed as there is no local "
                      "cr-lrp matching %s", ips, gateway_router)
            return

        if CONF.require_snat_disabled_for_tenant_networks:
            # Check if there is a SNAT entry for this LRP
            router = self.nb_idl.get_router(gateway_router)

            ips_without_snat = set(ips)
            for nat in router.nat:
                if nat.type == constants.OVN_SNAT:
                    net = ipaddress.ip_network(nat.logical_ip, strict=False)
                    for ip in list(ips_without_snat):
                        if ipaddress.ip_address(ip.split('/')[0]) in net:
                            ips_without_snat.discard(ip)

            if len(ips_without_snat) == 0:
                LOG.info('All ips (%s) were removed due to SNAT requirement '
                         'when exposing subnet %s for router %s', ips,
                         subnet_info['network'], gateway_router)
                return

            if len(set(ips)) != len(ips_without_snat):
                LOG.info('When exposing subnet %s for router %s, these ips '
                         'were removed for SNAT: %s', subnet_info['network'],
                         gateway_router, set(ips) - ips_without_snat)
                ips = list(ips_without_snat)

        try:
            self._expose_router_lsp(ips, subnet_info, cr_lrp_info)
        except (exceptions.ExposeDeniedForAddressScope,
                exceptions.WireFailure) as e:
            LOG.debug("Not exposing subnet CIDR's %s: %s", ips, e)
            return

        if (CONF.advertisement_method_tenant_networks ==
                constants.ADVERTISEMENT_METHOD_SUBNET):
            # Networks have been exposed via self._expose_router_lsp
            return

        ports = self.nb_idl.get_active_lsp(subnet_info['network'])
        for port in ports:
            # Check if the ip's on this port match the address scope. As the
            # port can be dual-stack, it could be that v4 is not allowed, but
            # v6 is allowed, so then only v6 address should be exposed.
            try:
                ips = self._ips_in_address_scope(
                    port_utils.get_ips_from_lsp(port),
                    subnet_info['address_scopes'])
            except exceptions.IpAddressNotFound:
                continue

            if not ips:
                # All ip's have been removed due to address scope requirement
                continue

            mac = port_utils.get_mac_from_lsp(port)
            ips_info = {
                'mac': mac,
                'cidrs': port.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                               "").split(),
                'type': port.type,
                'logical_switch': port.external_ids.get(
                    constants.OVN_LS_NAME_EXT_ID_KEY)
            }
            self._expose_remote_ip(ips, ips_info)

    def _withdraw_subnet(self, ips, subnet_info):
        gateway_router = subnet_info['associated_router']
        if not gateway_router:
            LOG.debug("Subnet CIDRs %s not withdrawn as there is no associated"
                      " router", ips)
            return
        cr_lrp_info = self.ovn_local_cr_lrps.get(gateway_router)
        if not cr_lrp_info:
            # NOTE(ltomasbo) there is a chance the cr-lrp just got moved
            # to this node but was not yet processed. In that case there
            # is no need to withdraw the network as it was not exposed here
            LOG.debug("Subnet CIDRs %s not withdrawn as there is no local "
                      "cr-lrp matching %s", ips, gateway_router)
            return

        try:
            self._withdraw_router_lsp(ips, subnet_info, cr_lrp_info)
        except (exceptions.ExposeDeniedForAddressScope,
                exceptions.UnwireFailure) as e:
            # Log a message, but silently continue, to make sure we have
            # it all withdrawn
            LOG.debug("Withdraw router lsp failure for CIDR's %s: %s", ips, e)

        if (CONF.advertisement_method_tenant_networks ==
                constants.ADVERTISEMENT_METHOD_SUBNET):
            # Expose the routes per prefix, rather than per port.
            return

        ports = self.nb_idl.get_active_lsp(subnet_info['network'])
        for port in ports:
            try:
                ips = port_utils.get_ips_from_lsp(port)
            except exceptions.IpAddressNotFound:
                continue
            mac = port_utils.get_mac_from_lsp(port)
            ips_info = {
                'mac': mac,
                'cidrs': port.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                               "").split(),
                'type': port.type,
                'logical_switch': port.external_ids.get(
                    constants.OVN_LS_NAME_EXT_ID_KEY)
            }
            self._withdraw_remote_ip(ips, ips_info)

    def _expose_router_lsp(self, ips, subnet_info, cr_lrp_info):
        '''Expose the tenant router ip address (cidr) for given router

        Will raise WireException if wire_lrp_port raises an exception or
        if it returns False

        Will raise ExposeDeniedForAddressScope if configured address scopes
        do not match the ones in configuration (if configured)
        (and execution should stop)
        '''
        if not self._expose_tenant_networks:
            return True

        if (CONF.advertisement_method_tenant_networks ==
                constants.ADVERTISEMENT_METHOD_SUBNET):
            # Fix ips to be the network address, instead of the lrp address
            # so we can advertise the entire subnet. This way a cleanup of
            # EVPN would not remove the incorrect entry as well.
            ips = driver_utils.get_prefixes_from_ips(ips)

        elif CONF.exposing_method == constants.EXPOSE_METHOD_VRF:
            # For evpn, we work with routes and since we are not exposing per
            # subnet, we need to remove the subnetmask part, so the ips are
            # exposed as a /32 or /128. Otherwise the cleanup would remove
            # the exposed route again, since 10.0.0.1/24 (a router ip) would
            # not match the kernel route of 10.0.0.0/24.
            ips = [ip.split('/')[0] for ip in ips]

        ips_to_process = []
        for ip in ips:
            if not CONF.expose_tenant_networks:
                # This means CONF.expose_ipv6_gua_tenant_networks is enabled
                if not driver_utils.is_ipv6_gua(ip):
                    continue

            ips_to_process.append(ip)

        if not ips_to_process:
            # Silently return, since there are no ip's left to process and
            # the address_scope has nothing to do with it.
            return True

        ips = self._ips_in_address_scope(ips_to_process,
                                         subnet_info['address_scopes'])
        if not ips:
            # All ip's failed address scope test, so stop processing this lsp
            raise exceptions.ExposeDeniedForAddressScope(
                addresses=','.join(ips_to_process),
                address_scopes=subnet_info['address_scopes'],
                configured_scopes=self.allowed_address_scopes,
            )

        for ip in ips:
            try:
                if wire_utils.wire_lrp_port(
                        self.ovn_routing_tables_routes, ip,
                        cr_lrp_info.get('bridge_device'),
                        cr_lrp_info.get('bridge_vlan'),
                        self.ovn_routing_tables, cr_lrp_info.get('ips')):

                    logical_switch = cr_lrp_info['provider_switch']
                    self._exposed_ips.setdefault(logical_switch, {}).update(
                        {ip: {
                            'bridge_device': cr_lrp_info.get('bridge_device'),
                            'bridge_vlan': cr_lrp_info.get('bridge_vlan'),
                            'via': cr_lrp_info.get('ips')}})

                    self.ovn_local_lrps.setdefault(
                        subnet_info['network'], set()).add(ip)
                else:
                    error_msg = ("Something happen while exposing the subnet"
                                 "and they have not been properly exposed")
                    raise exceptions.WireFailure(cidr=ip, message=error_msg)

            except Exception as e:
                raise exceptions.WireFailure(cidr=ip, message=str(e)) from e

        return True

    def _withdraw_router_lsp(self, ips, subnet_info, cr_lrp_info):
        '''Withdraw the tenant router ip address (cidr) for given router

        Will raise UnwireException if wire_lrp_port raises an exception or
        if it returns False

        Will raise ExposeDeniedForAddressScope if configured address scopes
        do not match the ones in configuration (if configured)
        (and execution should stop)
        '''
        if not self._expose_tenant_networks:
            return True

        if (CONF.advertisement_method_tenant_networks ==
                constants.ADVERTISEMENT_METHOD_SUBNET):
            # Fix ips to be the network address, instead of the lrp address
            # so we can withdraw the corrent subnet entry.
            ips = driver_utils.get_prefixes_from_ips(ips)

        elif CONF.exposing_method == constants.EXPOSE_METHOD_VRF:
            # For evpn, we work with routes and since we are not exposing per
            # subnet, we need to remove the subnetmask part, so the ips are
            # withdrawn as a /32 or /128.
            ips = [ip.split('/')[0] for ip in ips]

        ips_to_process = []
        for ip in ips:
            if (not CONF.expose_tenant_networks and
                    not driver_utils.is_ipv6_gua(ip)):
                # This means CONF.expose_ipv6_gua_tenant_networks is enabled
                continue

            ips_to_process.append(ip)

        if not ips_to_process:
            # Silently return, since there are no ip's left to process and
            # the address_scope has nothing to do with it.
            return True

        ips = self._ips_in_address_scope(ips_to_process,
                                         subnet_info['address_scopes'])
        if not ips:
            # All ip's failed address scope test, so stop processing this lsp
            raise exceptions.ExposeDeniedForAddressScope(
                addresses=','.join(ips_to_process),
                address_scopes=subnet_info['address_scopes'],
                configured_scopes=self.allowed_address_scopes,
            )

        for ip in ips:
            try:
                if wire_utils.unwire_lrp_port(
                        self.ovn_routing_tables_routes, ip,
                        cr_lrp_info.get('bridge_device'),
                        cr_lrp_info.get('bridge_vlan'),
                        self.ovn_routing_tables, cr_lrp_info.get('ips')):

                    logical_switch = cr_lrp_info['provider_switch']
                    self._exposed_ips.get(logical_switch, {}).pop(ip, None)
                else:
                    error_msg = ("Something happened while withdrawing subnet"
                                 "and they have not been properly removed")
                    raise exceptions.UnwireFailure(cidr=ip, message=error_msg)
            except Exception as e:
                raise exceptions.UnwireFailure(cidr=ip, message=str(e)) from e

        try:
            del self.ovn_local_lrps[subnet_info['network']]
        except KeyError:
            # Router port for subnet already cleanup
            pass

        return True

    @lockutils.synchronized('nbbgp')
    def expose_ovn_lb_vip(self, lb):
        self._expose_ovn_lb_vip(lb)

    def _expose_ovn_lb_vip(self, lb):
        vip_port = lb.external_ids.get(constants.OVN_LB_VIP_PORT_EXT_ID_KEY)
        vip_ip = lb.external_ids.get(constants.OVN_LB_VIP_IP_EXT_ID_KEY)
        vip_router = lb.external_ids[
            constants.OVN_LB_LR_REF_EXT_ID_KEY].replace('neutron-', "", 1)
        vip_lsp = self.nb_idl.lsp_get(vip_port).execute(check_error=True)
        if not vip_lsp:
            LOG.debug("Something went wrong, VIP port %s not found", vip_port)
            return
        vip_net = vip_lsp.external_ids.get(constants.OVN_LS_NAME_EXT_ID_KEY)
        if vip_net in self.ovn_local_lrps.keys():
            # It is a VIP on a tenant network
            # NOTE: the LB is exposed through the cr-lrp, so we add the
            # vip_router instead of the logical switch
            ips_info = {'logical_switch': vip_router}
            self._expose_remote_ip([vip_ip], ips_info)
        else:
            # It is a VIP on a provider network
            localnet, bridge_device, bridge_vlan = self._get_provider_ls_info(
                vip_net)
            self._expose_provider_port([vip_ip], None, vip_net, bridge_device,
                                       bridge_vlan, localnet)

    @lockutils.synchronized('nbbgp')
    def withdraw_ovn_lb_vip(self, lb):
        self._withdraw_ovn_lb_vip(lb)

    def _withdraw_ovn_lb_vip(self, lb):
        vip_ip = lb.external_ids.get(constants.OVN_LB_VIP_IP_EXT_ID_KEY)
        vip_router = lb.external_ids[
            constants.OVN_LB_LR_REF_EXT_ID_KEY].replace('neutron-', "", 1)

        cr_lrp_info = self.ovn_local_cr_lrps.get(vip_router)
        if not cr_lrp_info:
            return
        provider_ls = cr_lrp_info['provider_switch']
        if self._exposed_ips.get(provider_ls, {}).get(vip_ip):
            # VIP is on provider network
            self._withdraw_provider_port([vip_ip],
                                         cr_lrp_info['provider_switch'],
                                         cr_lrp_info['bridge_device'],
                                         cr_lrp_info['bridge_vlan'])
        else:
            # VIP is on tenant network
            ips_info = {'logical_switch': vip_router}
            self._withdraw_remote_ip([vip_ip], ips_info)

    @lockutils.synchronized('nbbgp')
    def expose_ovn_lb_fip(self, lb):
        self._expose_ovn_lb_fip(lb)

    def _expose_ovn_lb_fip(self, lb):
        vip_port = lb.external_ids.get(constants.OVN_LB_VIP_PORT_EXT_ID_KEY)
        vip_lsp = self.nb_idl.lsp_get(vip_port).execute(check_error=True)
        if not vip_lsp:
            LOG.debug("Something went wrong, VIP port %s not found", vip_port)
            return

        try:
            external_ip, external_mac, ls_name = (
                self.get_port_external_ip_and_ls(vip_lsp.name))
        except nb_exceptions.NATNotFound as e:
            LOG.debug("Something went wrong with the VIP %s: %s",
                      vip_port, e)
            return
        self._expose_fip(external_ip, external_mac, ls_name, vip_lsp)

    def _get_parameters_from_lb(self, lb, include_mac_and_localnet=False):
        for fipport in lb.vips.keys():
            fip, port = fipport.split(':')
            break
        else:
            return

        router = lb.external_ids.get(
            constants.OVN_LR_NAME_EXT_ID_KEY, '').replace('neutron-', "", 1)
        if not router:
            return
        cr_lrp_info = self.ovn_local_cr_lrps.get(router)
        if not cr_lrp_info:
            return
        net, bridge_device, bridge_vlan = self._get_ls_localnet_info(
            cr_lrp_info['provider_switch'])
        kwargs = {
            'port_ips': [fip],
            'logical_switch': cr_lrp_info['provider_switch'],
            'bridge_device': bridge_device,
            'bridge_vlan': bridge_vlan}

        if include_mac_and_localnet:
            kwargs['mac'] = None
            kwargs['localnet'] = net

        return kwargs

    @lockutils.synchronized('nbbgp')
    def expose_ovn_pf_lb_fip(self, lb):
        self._expose_ovn_pf_lb_fip(lb)

    @lockutils.synchronized('nbbgp')
    def withdraw_ovn_pf_lb_fip(self, lb):
        self._withdraw_ovn_pf_lb_fip(lb)

    def _withdraw_ovn_pf_lb_fip(self, lb):
        kwargs = self._get_parameters_from_lb(lb)
        self._withdraw_provider_port(**kwargs) if kwargs else None

    def _expose_ovn_pf_lb_fip(self, lb):
        kwargs = self._get_parameters_from_lb(lb, True)
        self._expose_provider_port(**kwargs) if kwargs else None

    @lockutils.synchronized('nbbgp')
    def withdraw_ovn_lb_fip(self, lb):
        self._withdraw_ovn_lb_fip(lb)

    def _withdraw_ovn_lb_fip(self, lb):
        vip_fip = lb.external_ids.get(constants.OVN_LB_VIP_FIP_EXT_ID_KEY)
        # OVN loadbalancers ARPs are replied by router port
        vip_router = lb.external_ids.get(
            constants.OVN_LB_LR_REF_EXT_ID_KEY, "").replace('neutron-', "", 1)
        if not vip_router:
            return
        cr_lrp_info = self.ovn_local_cr_lrps.get(vip_router)
        if not cr_lrp_info:
            return
        self._withdraw_provider_port([vip_fip],
                                     cr_lrp_info['provider_switch'],
                                     cr_lrp_info['bridge_device'],
                                     cr_lrp_info['bridge_vlan'])

    def _ips_in_address_scope(self, ips, address_scopes):
        return [ip
                for ip in ips
                if self._address_scope_allowed(ip, address_scopes)]

    def _address_scope_allowed(self, ip, address_scopes):
        if not self.allowed_address_scopes:
            # No address scopes to filter on => announce everything
            return True

        # if we should filter on address scopes and this port has no
        # address scopes set we do not need to expose it
        if not any(address_scopes.values()):
            return False
        # if address scope does not match, no need to expose it
        ip_version = linux_net.get_ip_version(ip)

        return address_scopes[ip_version] in self.allowed_address_scopes



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import port as port_utils
from ovn_bgp_agent.drivers.openstack.watchers import base_watcher
from ovn_bgp_agent.utils import helpers

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
_SYNC_STATE_LOCK = lockutils.ReaderWriterLock()


class PortBindingChassisCreatedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(PortBindingChassisCreatedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.mac[0]):
                return False
            if not bool(row.up[0]):
                return False

            if row.chassis[0].name != self.agent.chassis:
                return False
            if hasattr(old, 'chassis'):
                if not old.chassis or row.chassis != old.chassis:
                    return True
            if hasattr(old, 'up'):
                if not bool(old.up[0]):
                    return True
        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type not in constants.OVN_VIF_PORT_TYPES:
            return
        with _SYNC_STATE_LOCK.read_lock():
            ips = row.mac[0].split(' ')[1:]
            self.agent.expose_ip(ips, row)


class PortBindingChassisDeletedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE, self.ROW_DELETE,)
        super(PortBindingChassisDeletedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.mac[0]):
                return False
            if event == self.ROW_DELETE:
                return row.chassis[0].name == self.agent.chassis

            if hasattr(old, 'chassis'):
                if (old.chassis[0].name == self.agent.chassis and
                        (not row.chassis or row.chassis != old.chassis)):
                    return True
            if hasattr(old, 'up'):
                # this requires to have unchanged chassis and being the local
                # one. If there was a chassis change, then it was already
                # processed before
                if (row.chassis[0].name == self.agent.chassis and
                        bool(old.up[0]) and not bool(row.up[0])):
                    return True
        except (IndexError, AttributeError):
            return False
        return False

    def _run(self, event, row, old):
        if row.type not in constants.OVN_VIF_PORT_TYPES:
            return
        with _SYNC_STATE_LOCK.read_lock():
            ips = row.mac[0].split(' ')[1:]
            self.agent.withdraw_ip(ips, row)


class FIPSetEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(FIPSetEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            return (not row.chassis and
                    row.nat_addresses != old.nat_addresses and
                    not row.logical_port.startswith('lrp-'))
        except (AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type != constants.OVN_PATCH_VIF_PORT_TYPE:
            return
        with _SYNC_STATE_LOCK.read_lock():
            # NOTE(ltomasbo): nat_addresses has the same format, where
            # different IPs can be present:
            # ["fa:16:3e:77:7f:9c 172.24.100.229 172.24.100.112
            #  is_chassis_resident(\"
            #      cr-lrp-add962d2-21ab-4733-b6ef-35538eff25a8\")"]
            old_cr_lrps = {}
            for nat in old.nat_addresses:
                ips = nat.strip().split(" ")[1:-1]
                port = nat.strip().split(" ")[-1].split("\"")[1]
                old_cr_lrps.setdefault(port, set()).update(ips)
            for nat in row.nat_addresses:
                ips = nat.strip().split(" ")[1:-1]
                port = nat.strip().split(" ")[-1].split("\"")[1]
                ips_to_expose = [ip for ip in ips
                                 if ip not in old_cr_lrps.get(port, set())]
                if ips_to_expose:
                    self.agent.expose_ip(ips_to_expose, row,
                                         associated_port=port)


class FIPUnsetEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(FIPUnsetEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            return (not row.chassis and
                    row.nat_addresses != old.nat_addresses and
                    not row.logical_port.startswith('lrp-'))
        except (AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type != constants.OVN_PATCH_VIF_PORT_TYPE:
            return
        with _SYNC_STATE_LOCK.read_lock():
            # NOTE(ltomasbo): nat_addresses has the same format, where
            # different IPs can be present:
            # ["fa:16:3e:77:7f:9c 172.24.100.229 172.24.100.112
            #  is_chassis_resident(\"
            #      cr-lrp-add962d2-21ab-4733-b6ef-35538eff25a8\")"]
            current_cr_lrps = {}
            for nat in row.nat_addresses:
                ips = nat.strip().split(" ")[1:-1]
                port = nat.strip().split(" ")[-1].split("\"")[1]
                current_cr_lrps.setdefault(port, set()).update(ips)
            for nat in old.nat_addresses:
                ips = nat.strip().split(" ")[1:-1]
                port = nat.strip().split(" ")[-1].split("\"")[1]
                ips_to_withdraw = [ip for ip in ips
                                   if ip not in current_cr_lrps.get(port,
                                                                    set())]
                if ips_to_withdraw:
                    self.agent.withdraw_ip(ips_to_withdraw, row,
                                           associated_port=port)


class SubnetRouterAttachedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_CREATE,)
        super(SubnetRouterAttachedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.mac[0]):
                return False
            return (not row.chassis and
                    row.logical_port.startswith('lrp-') and
                    "chassis-redirect-port" not in row.options.keys())
        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type != constants.OVN_PATCH_VIF_PORT_TYPE:
            return
        with _SYNC_STATE_LOCK.read_lock():
            ip_address = row.mac[0].split(' ')[1]
            self.agent.expose_subnet(ip_address, row)


class SubnetRouterUpdateEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(SubnetRouterUpdateEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        # This will match if the mac field has changed between old and row.
        # This can happen when you have multiple subnets in the same network,
        # those will be added/removed to/from the same lrp-port in the mac
        # field.
        # Format:
        # mac = [ff:ff:ff:ff:ff:ff subnet1/cidr subnet2/cidr [...]]
        try:
            # single and dual-stack format
            if (not port_utils.has_ip_address_defined(row.mac[0]) and
                    not port_utils.has_ip_address_defined(old.mac[0])):
                return False
            return (
                not row.chassis and
                row.logical_port.startswith("lrp-") and
                "chassis-redirect-port" not in row.options.keys() and
                old.mac != row.mac
            )
        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type != constants.OVN_PATCH_VIF_PORT_TYPE:
            return
        with _SYNC_STATE_LOCK.read_lock():
            self.agent.update_subnet(old, row)


class SubnetRouterDetachedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_DELETE,)
        super(SubnetRouterDetachedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.mac[0]):
                return False
            return (not row.chassis and
                    row.logical_port.startswith('lrp-') and
                    "chassis-redirect-port" not in row.options.keys())
        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type != constants.OVN_PATCH_VIF_PORT_TYPE:
            return
        with _SYNC_STATE_LOCK.read_lock():
            ip_address = row.mac[0].split(' ')[1]
            self.agent.withdraw_subnet(ip_address, row)


class TenantPortCreatedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(TenantPortCreatedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            # Handling the case for unknown MACs when configdrive is used
            # instead of dhcp
            if row.mac == ['unknown']:
                n_cidrs = row.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                               "").split()
                if not n_cidrs:
                    return False
            # single and dual-stack format
            elif not port_utils.has_ip_address_defined(row.mac[0]):
                return False
            return (not old.chassis and row.chassis and
                    self.agent.ovn_local_lrps != [])
        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type not in (constants.OVN_VM_VIF_PORT_TYPE,
                            constants.OVN_VIRTUAL_VIF_PORT_TYPE):
            return
        with _SYNC_STATE_LOCK.read_lock():
            if row.mac == ['unknown']:
                # Handling the case for unknown MACs when configdrive is used
                # instead of dhcp
                n_cidrs = row.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                               "")
                ips = [ip.split("/")[0] for ip in n_cidrs.split(" ")]
            else:
                ips = row.mac[0].split(' ')[1:]
            self.agent.expose_remote_ip(ips, row)


class TenantPortDeletedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE, self.ROW_DELETE,)
        super(TenantPortDeletedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            if row.mac == ['unknown']:
                # Handling the case for unknown MACs when configdrive is used
                # instead of dhcp
                n_cidrs = row.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                               "").split()
                if not n_cidrs:
                    return False
            # single and dual-stack format
            elif not port_utils.has_ip_address_defined(row.mac[0]):
                return False
            if event == self.ROW_UPDATE:
                return (old.chassis and not row.chassis and
                        self.agent.ovn_local_lrps != [])
            if event == self.ROW_DELETE:
                return (self.agent.ovn_local_lrps != [])
        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type not in (constants.OVN_VM_VIF_PORT_TYPE,
                            constants.OVN_VIRTUAL_VIF_PORT_TYPE):
            return
        if event == self.ROW_UPDATE:
            chassis = old.chassis
        else:
            chassis = row.chassis
        with _SYNC_STATE_LOCK.read_lock():
            if row.mac == ['unknown']:
                # Handling the case for unknown MACs when configdrive is used
                # instead of dhcp
                n_cidrs = row.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                               "")
                ips = [ip.split("/")[0] for ip in n_cidrs.split(" ")]
            else:
                ips = row.mac[0].split(' ')[1:]
            self.agent.withdraw_remote_ip(ips, row, chassis)


class OVNLBVIPPortEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_CREATE, self.ROW_DELETE,)
        super(OVNLBVIPPortEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        # The ovn lb balancers are exposed through the cr-lrp, so if the
        # local agent does not have any cr-lrp associated there is no need
        # to process the event
        try:
            # it should not have mac, no chassis, and status down
            if not row.mac and not row.chassis and not row.up[0]:
                return bool(self.agent.ovn_local_cr_lrps)
            return False
        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type != constants.OVN_VM_VIF_PORT_TYPE:
            return

        with _SYNC_STATE_LOCK.read_lock():
            # This is depending on the external-id information added by
            # neutron, regarding the neutron:cidrs
            ext_n_cidr = row.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                              "")
            if not ext_n_cidr:
                return

            ovn_lb_ip = ext_n_cidr.strip().split(" ")[0].split("/")[0]
            if event == self.ROW_DELETE:
                self.agent.withdraw_ovn_lb(ovn_lb_ip, row)
            if event == self.ROW_CREATE:
                self.agent.expose_ovn_lb(ovn_lb_ip, row)


class OVNLBMemberCreateEvent(base_watcher.OVNLBEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_CREATE,)
        super(OVNLBMemberCreateEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        # Only process event if the local node has a cr-lrp ports associated
        return bool(self.agent.ovn_local_cr_lrps)

    def _run(self, event, row, old):
        # Only process event if the local node has a cr-lrp port whose provider
        # datapath is included into the loadbalancer. This means the
        # loadbalancer has the VIP on a provider network
        # Also, the cr-lrp port needs to have subnet datapaths (LS) associated
        # to it that include the load balancer
        if not self.agent.ovn_local_cr_lrps:
            return
        try:
            row_dps = row.datapaths
        except AttributeError:
            row_dps = []

        row_dps, router_dps = helpers.get_lb_datapaths(row)
        if not row_dps:
            # No need to continue. There is no need to expose it as there is
            # no datapaths (aka members).
            return
        vip_port = self.agent.sb_idl.get_ovn_vip_port(row.name)
        if not vip_port:
            return
        vip_ip = vip_port.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY, "")
        if not vip_ip:
            return
        vip_ip = vip_ip.strip().split(" ")[0].split("/")[0]
        associated_cr_lrp_port = None

        if not router_dps and not (CONF.expose_tenant_networks or
                                   CONF.expose_ipv6_gua_tenant_networks):
            # assume all the members are connected through the same router
            # so only one member needs to be checked
            member_dp = row_dps[0]
            # get lrps on that dp (patch ports)
            router_lrps = (
                self.agent.sb_idl.get_lrps_for_datapath(member_dp))
            for lrp in router_lrps:
                router_dps.append(self.agent.sb_idl.get_port_datapath(lrp))

        for cr_lrp_port, cr_lrp_info in self.agent.ovn_local_cr_lrps.items():
            if vip_port.datapath != cr_lrp_info.get('provider_datapath'):
                continue
            if cr_lrp_info.get('subnets_datapath'):
                if set(row_dps).intersection(set(
                        cr_lrp_info.get('subnets_datapath').values())):
                    associated_cr_lrp_port = cr_lrp_port
                    break
            else:
                if cr_lrp_info.get('router_datapath') in router_dps:
                    associated_cr_lrp_port = cr_lrp_port
                    break
        else:
            return

        with _SYNC_STATE_LOCK.read_lock():
            return self.agent.expose_ovn_lb_on_provider(
                vip_ip, row.name, associated_cr_lrp_port)


class OVNLBMemberDeleteEvent(base_watcher.OVNLBEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_DELETE,)
        super(OVNLBMemberDeleteEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        # Only process event if the local node has the lb exported
        return bool(self.agent.provider_ovn_lbs.get(row.name))

    def _run(self, event, row, old):
        associated_cr_lrp_port = self.agent.provider_ovn_lbs[row.name].get(
            'gateway_port')
        if not associated_cr_lrp_port:
            # Something is wrong, not enough information to proceed
            return
        with _SYNC_STATE_LOCK.read_lock():
            # loadbalancer deleted. Withdraw the VIP through the cr-lrp
            return self.agent.withdraw_ovn_lb_on_provider(
                row.name, associated_cr_lrp_port)


class LocalnetCreateDeleteEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_CREATE, self.ROW_DELETE,)
        super(LocalnetCreateDeleteEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        return row.type == constants.OVN_LOCALNET_VIF_PORT_TYPE

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            self.agent.sync()



# Copyright 2024 team.blue/nl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import driver_utils
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests import utils


class TestDriverUtils(test_base.TestCase):

    def setUp(self):
        super(TestDriverUtils, self).setUp()

    def test_is_ipv6_gua(self):
        self.assertFalse(driver_utils.is_ipv6_gua('1.1.1.1'))
        self.assertFalse(driver_utils.is_ipv6_gua('fe80::1337'))
        self.assertTrue(driver_utils.is_ipv6_gua('2a01:db8::1337'))

    def test_get_addr_scopes(self):
        subnet_pool_addr_scope4 = '88e8aec3-da29-402d-becf-9fa2c38e69b8'
        subnet_pool_addr_scope6 = 'b7834aeb-2aa2-40ac-a8b5-2cded713cb58'

        # Both address pools set
        port = utils.create_row(external_ids={
            constants.SUBNET_POOL_ADDR_SCOPE4: subnet_pool_addr_scope4,
            constants.SUBNET_POOL_ADDR_SCOPE6: subnet_pool_addr_scope6,
        })
        self.assertDictEqual(driver_utils.get_addr_scopes(port), {
            constants.IP_VERSION_4: subnet_pool_addr_scope4,
            constants.IP_VERSION_6: subnet_pool_addr_scope6,
        })

        # Only IPv4
        port = utils.create_row(external_ids={
            constants.SUBNET_POOL_ADDR_SCOPE4: subnet_pool_addr_scope4,
        })
        self.assertDictEqual(driver_utils.get_addr_scopes(port), {
            constants.IP_VERSION_4: subnet_pool_addr_scope4,
            constants.IP_VERSION_6: None,
        })

        # Only IPv6
        port = utils.create_row(external_ids={
            constants.SUBNET_POOL_ADDR_SCOPE6: subnet_pool_addr_scope6,
        })
        self.assertDictEqual(driver_utils.get_addr_scopes(port), {
            constants.IP_VERSION_4: None,
            constants.IP_VERSION_6: subnet_pool_addr_scope6,
        })

        # No Address pools
        port = utils.create_row(external_ids={})
        self.assertDictEqual(driver_utils.get_addr_scopes(port), {
            constants.IP_VERSION_4: None,
            constants.IP_VERSION_6: None,
        })

    def test_get_port_chassis_from_options(self):
        my_host = 'foo-host'

        # it is a VM port type, should use options field.
        row = utils.create_row(
            external_ids={constants.OVN_HOST_ID_EXT_ID_KEY: 'bar-host'},
            options={constants.OVN_REQUESTED_CHASSIS: my_host})

        self.assertEqual(driver_utils.get_port_chassis(row, chassis=my_host),
                         my_host)

    def test_get_port_chassis_from_external_ids(self):
        my_host = 'foo-host'

        # it is a VM port type, should use options field.
        row = utils.create_row(
            external_ids={constants.OVN_HOST_ID_EXT_ID_KEY: my_host})

        self.assertEqual(driver_utils.get_port_chassis(row, chassis=my_host),
                         my_host)

    def test_get_port_chassis_from_external_ids_virtual_port(self):
        my_host = 'foo-host'

        # it is a VM port type, should use options field.
        row = utils.create_row(
            external_ids={constants.OVN_HOST_ID_EXT_ID_KEY: my_host},
            options={constants.OVN_REQUESTED_CHASSIS: 'bar-host'},
            type=constants.OVN_VIRTUAL_VIF_PORT_TYPE)

        self.assertEqual(driver_utils.get_port_chassis(row, chassis=my_host),
                         my_host)

    def test_get_port_chassis_no_information(self):
        row = utils.create_row()
        self.assertEqual(driver_utils.get_port_chassis(row, chassis='foo'),
                         None)

    def test_check_name_prefix(self):
        lb = utils.create_row(name='some-name')
        self.assertTrue(driver_utils.check_name_prefix(lb, 'some'))
        self.assertFalse(driver_utils.check_name_prefix(lb, 'other'))

        lb = utils.create_row(no_name='aa')
        self.assertFalse(driver_utils.check_name_prefix(lb, ''))

    def is_pf_lb(self):
        lb = utils.create_row(name='pf-floating-ip-someuuid')
        self.assertTrue(driver_utils.is_pf_lb(lb))

        lb = utils.create_row(name='lb-someuuid')
        self.assertFalse(driver_utils.is_pf_lb(lb))

    def test_ips_per_version(self):
        ips = ['10.0.0.1/32', 'fe80::1/128']
        self.assertDictEqual(driver_utils.ips_per_version(ips), {
            constants.IP_VERSION_4: '10.0.0.1',
            constants.IP_VERSION_6: 'fe80::1',
        })

        # More than 1 ip per version and without prefix
        ips = ['10.0.0.1/32', '10.0.0.254', 'fe80::1', 'fc00::1/128']
        self.assertDictEqual(driver_utils.ips_per_version(ips), {
            constants.IP_VERSION_4: '10.0.0.254',
            constants.IP_VERSION_6: 'fc00::1',
        })

        # Only IPv4
        ips = ['10.0.0.1/32']
        self.assertDictEqual(driver_utils.ips_per_version(ips), {
            constants.IP_VERSION_4: '10.0.0.1',
            constants.IP_VERSION_6: None,
        })

        # Only IPv6
        ips = ['fc00::1/128']
        self.assertDictEqual(driver_utils.ips_per_version(ips), {
            constants.IP_VERSION_4: None,
            constants.IP_VERSION_6: 'fc00::1',
        })

    def test_get_prefixes_from_ips(self):
        # IPv4
        ips = ['192.168.0.1/24', '192.168.0.244/28', '172.13.37.59/27']
        self.assertListEqual(driver_utils.get_prefixes_from_ips(ips),
                             ['192.168.0.0/24', '192.168.0.240/28',
                              '172.13.37.32/27'])

        # IPv6
        ips = ['fe80::5097/64', 'ff00::13:37/112', 'fc00::1/46']
        self.assertListEqual(driver_utils.get_prefixes_from_ips(ips),
                             ['fe80::/64', 'ff00::13:0/112', 'fc00::/46'])

        # combined.
        ips = ['172.13.37.59/27', 'ff00::13:37/112']
        self.assertListEqual(driver_utils.get_prefixes_from_ips(ips),
                             ['172.13.37.32/27', 'ff00::13:0/112'])

    def test_get_port_vlan_untagged(self):
        port = utils.create_row(tag=[])
        self.assertEqual('0', driver_utils.get_port_vlan(port))

    def test_get_port_vlan_tagged(self):
        port = utils.create_row(tag=[100])
        self.assertEqual('100', driver_utils.get_port_vlan(port))

        port = utils.create_row(tag=[100, 123])
        self.assertEqual('100', driver_utils.get_port_vlan(port))

    def test_get_port_vrf_settings(self):
        row = utils.create_row(external_ids={
            constants.OVN_EVPN_TYPE_EXT_ID_KEY: constants.OVN_EVPN_TYPE_L3,
            constants.OVN_EVPN_VNI_EXT_ID_KEY: 1001,
        })
        self.assertEqual('l3::1001', driver_utils.get_port_vrf_settings(row))

    def test_get_port_vrf_settings_no_vni(self):
        row = utils.create_row(external_ids={
            constants.OVN_EVPN_TYPE_EXT_ID_KEY: constants.OVN_EVPN_TYPE_L3,
        })
        self.assertEqual('', driver_utils.get_port_vrf_settings(row))

    def test_get_port_vrf_settings_no_type(self):
        row = utils.create_row(external_ids={
            constants.OVN_EVPN_VNI_EXT_ID_KEY: 1001,
        })
        self.assertEqual('', driver_utils.get_port_vrf_settings(row))

    def test_get_port_vrf_settings_not_provided(self):
        row = utils.create_row()
        self.assertIsNone(driver_utils.get_port_vrf_settings(row))



# Copyright 2024 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import port
from ovn_bgp_agent import exceptions
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests import utils as test_utils


class TestHasIpAddressDefined(test_base.TestCase):
    def test_no_ip_address(self):
        self.assertFalse(
            port.has_ip_address_defined('aa:bb:cc:dd:ee:ff'))

    def test_one_ip_address(self):
        self.assertTrue(
            port.has_ip_address_defined('aa:bb:cc:dd:ee:ff 10.10.1.16'))

    def test_two_ip_addresses(self):
        self.assertTrue(
            port.has_ip_address_defined(
                'aa:bb:cc:dd:ee:ff 10.10.1.16 10.10.1.17'))

    def test_three_ip_addresses(self):
        self.assertTrue(
            port.has_ip_address_defined(
                'aa:bb:cc:dd:ee:ff 10.10.1.16 10.10.1.17 10.10.1.18'))


class TestGetFip(test_base.TestCase):
    def test_get_fip(self):
        value = 'foo'
        lsp = test_utils.create_row(
            external_ids={constants.OVN_FIP_EXT_ID_KEY: value})
        observed = port.get_fip(lsp)

        self.assertEqual(value, observed)

    def test_get_fip_not_present(self):
        lsp = test_utils.create_row()
        self.assertIsNone(port.get_fip(lsp))


class TestHasAdditionalBinding(test_base.TestCase):
    def test_has_multiple_chassis(self):
        lsp = test_utils.create_row(
            options={constants.OVN_REQUESTED_CHASSIS: "chassis1,chassis2"})

        self.assertTrue(port.has_additional_binding(lsp))

    def test_has_one_chassis(self):
        lsp = test_utils.create_row(
            options={constants.OVN_REQUESTED_CHASSIS: "chassis1"})

        self.assertFalse(port.has_additional_binding(lsp))

    def test_no_options(self):
        lsp = test_utils.create_row()

        self.assertFalse(port.has_additional_binding(lsp))

    def test_no_requested_chassis(self):
        lsp = test_utils.create_row(
            options={})

        self.assertFalse(port.has_additional_binding(lsp))

    def test_empty_requested_chassis(self):
        lsp = test_utils.create_row(
            options={constants.OVN_REQUESTED_CHASSIS: ""})

        self.assertFalse(port.has_additional_binding(lsp))


class TestGetAddressList(test_base.TestCase):
    def test_get_list(self):
        expected_list = ["mac", "ip1", "ip2"]
        lsp = test_utils.create_row(
            addresses=["mac ip1 ip2"])
        observed = port.get_address_list(lsp)

        self.assertListEqual(expected_list, observed)

    def test_get_list_strip(self):
        expected_list = ["mac", "ip1", "ip2"]
        lsp = test_utils.create_row(
            addresses=["  mac ip1 ip2 "])
        observed = port.get_address_list(lsp)

        self.assertListEqual(expected_list, observed)

    def test_get_list_no_addresses(self):
        lsp = test_utils.create_row()
        observed = port.get_address_list(lsp)

        self.assertListEqual([], observed)

    def test_get_list_empty(self):
        lsp = test_utils.create_row(addresses=[])
        observed = port.get_address_list(lsp)

        self.assertListEqual([], observed)

    def test_get_list_empty_string(self):
        lsp = test_utils.create_row(addresses=[""])
        observed = port.get_address_list(lsp)

        self.assertListEqual([], observed)


class TestGetMacFromLsp(test_base.TestCase):
    def test_get_mac(self):
        mac = 'mac'
        lsp = test_utils.create_row(
            addresses=["%s ip1 ip2" % mac])
        observed_mac = port.get_mac_from_lsp(lsp)

        self.assertEqual(mac, observed_mac)

    def test_get_mac_empty_list(self):
        lsp = test_utils.create_row(
            addresses=[])
        self.assertRaises(
            exceptions.MacAddressNotFound, port.get_mac_from_lsp, lsp)


class TestGetIpsFromLsp(test_base.TestCase):
    def test_get_ips(self):
        ips = ['ip1', 'ip2']
        lsp = test_utils.create_row(
            addresses=["mac " + ' '.join(ips)])
        observed_ips = port.get_ips_from_lsp(lsp)

        self.assertEqual(ips, observed_ips)

    def test_get_ips_empty_list(self):
        lsp = test_utils.create_row(
            addresses=[])
        self.assertRaises(
            exceptions.IpAddressNotFound, port.get_ips_from_lsp, lsp)
